from __future__ import print_function,  absolute_import
import underworld as uw
from UWGeodynamics import nd
from scipy.interpolate import interp1d

class FreeSurfaceProcessor(object):
    """FreeSurfaceProcessor"""

    def __init__(self, Model):
        """__init__

        Parameters
        ----------

        Model : UWGeodynamics Model

        """
        self.Model = Model

        # Create the tools
        self.TField = self.Model.mesh.add_variable( nodeDofCount=1 )
        self.TField.data[:, 0] = self.Model.mesh.data[:, 1]

        self.top = self.Model.top_wall
        self.bottom = self.Model.bottom_wall

        # Create boundary condition
        self._conditions = uw.conditions.DirichletCondition(
            variable=self.TField,
            indexSetsPerDof=(self.top + self.bottom,))

        # Create Eq System
        self._system = uw.systems.SteadyStateHeat(temperatureField = self.TField,
                                                  fn_diffusivity = 1.0,
                                                  conditions = self._conditions)

        self._solver = uw.systems.Solver(self._system)

    def _solve_sle(self):
        self._solver.solve()

    def _advect_surface(self, dt):

        if self.top:
            # Extract top surface
            x = self.Model.mesh.data[self.top.data][:, 0]
            y = self.Model.mesh.data[self.top.data][:, 1]

            # Extract velocities from top
            vx = self.Model.velocityField.data[self.top.data][:,0]
            vy = self.Model.velocityField.data[self.top.data][:,1]

            # Advect top surface
            x2 = x + vx * nd(dt)
            y2 = y + vy * nd(dt)

            # Spline top surface
            f = interp1d(x2, y2, kind='cubic', fill_value='extrapolate')

            self.TField.data[self.top.data, 0] = f(x)
        uw.barrier()
        self.TField.syncronise()

    def _update_mesh(self):

        with self.Model.mesh.deform_mesh():
            # Last dimension is the vertical dimension
            self.Model.mesh.data[:, -1] = self.TField.data[:, 0]

    def solve(self, dt):

        # First we advect the surface
        self._advect_surface(dt)
        # Then we solve the system of linear equation
        self._solve_sle()
        # Finally we update the mesh
        self._update_mesh()



from __future__ import print_function,  absolute_import
import underworld as uw
import h5py
import numpy as np
import os
from mpi4py import MPI
from UWGeodynamics.scaling import Dimensionalize
from UWGeodynamics.scaling import nonDimensionalize
from UWGeodynamics.scaling import UnitRegistry as u
from UWGeodynamics.version import git_revision as __git_revision__


class MeshVariable(uw.mesh.MeshVariable):

    def __init__(self, mesh, nodeDofCount, dataType="double", **kwargs):
        super(MeshVariable, self).__init__(mesh, nodeDofCount, dataType, **kwargs)

    def load(self, filename, interpolate=False):
        """
        Load the MeshVariable from disk.

        Parameters
        ----------
        filename: str
            The filename for the saved file. Relative or absolute paths may be
            used, but all directories must exist.
        interpolate: bool
            Set to True to interpolate a file containing different resolution data.
            Note that a temporary MeshVariable with the file data will be build
            on **each** processor. Also note that the temporary MeshVariable
            can only be built if its corresponding mesh file is available.
            Also note that the supporting mesh mush be regular.

        Notes
        -----
        This method must be called collectively by all processes.

        If the file data array is the same length as the current variable
        global size, it is assumed the file contains compatible data. Note that
        this may not be the case where for example where you have saved using a
        2*8 resolution mesh, then loaded using an 8*2 resolution mesh.

        Provided files must be in hdf5 format, and use the correct schema.

        Example
        -------
        Refer to example provided for 'save' method.

        """
        if not isinstance(filename, str):
            raise TypeError("Expected filename to be provided as a string")

        # get field and mesh information
        h5f = h5py.File( filename, "r", driver='mpio', comm=MPI.COMM_WORLD );
        dset = h5f.get('data')

        # get units
        try:
            units = h5f.attrs["units"]
        except KeyError:
            units = None

        if units and units != "None":
            units = u.parse_expression(units)
        else:
            units = None

        if dset == None:
            raise RuntimeError("Can't find the 'data' in hdf5 file '{0}'".format(filename) )

        dof = dset.shape[1]
        if dof != self.data.shape[1]:
            raise RuntimeError("Can't load hdf5 '{0}', incompatible data shape".format(filename))

        if len(dset) == self.mesh.nodesGlobal:

            # assume dset matches field exactly
            mesh = self.mesh
            local = mesh.nodesLocal

            with dset.collective:
                self.data[0:local] = dset[mesh.data_nodegId[0:local],:]

        else:
            if not interpolate:
                raise RuntimeError("Provided data file appears to be for a different resolution MeshVariable.\n"\
                                   "If you would like to interpolate the data to the current variable, please set\n" \
                                   "the 'interpolate' parameter. Check docstring for important caveats of interpolation method.")

            # if here then we build a local version of the entire file field and interpolate it's values

            # first get file field's mesh
            if h5f.get('mesh') == None:
                raise RuntimeError("The hdf5 field to be loaded with interpolation must have an associated "+
                        "'mesh' hdf5 file. Resave the field with its associated mesh."+
                        "i.e. myField.save(\"filename.h5\", meshFilename)" )
            # get resolution of old mesh
            res = h5f['mesh'].attrs.get('mesh resolution')
            if res is None:
                raise RuntimeError("Can't read the 'mesh resolution' for the field hdf5 file,"+
                       " was it created correctly?")

            # get max of old mesh
            inputMax = h5f['mesh'].attrs.get('max')
            if inputMax is None:
                raise RuntimeError("Can't read the 'max' for the field hdf5 file,"+
                       " was it created correctly?")

            inputMin = h5f['mesh'].attrs.get('min')
            if inputMin is None:
                raise RuntimeError("Can't read the 'min' for the field hdf5 file,"+
                       " was it created correctly?")
            regular = h5f['mesh'].attrs.get('regular')
            if regular and regular!=True:
                raise RuntimeError("Saved mesh file appears to correspond to a irregular mesh.\n"\
                                   "Interpolating from irregular mesh not currently supported." )

            elType = h5f['mesh'].attrs.get('elementType')
            # for backwards compatiblity, the 'elementType' attribute was added Feb2017
            if elType == None:
                elType = 'Q1'

            # build the NON-PARALLEL field and mesh
            inputMesh = uw.mesh.FeMesh_Cartesian( elementType = (elType+"/DQ0"), # only geometryMesh can be saved
                                          elementRes  = tuple(res),
                                          minCoord    = tuple(inputMin),
                                          maxCoord    = tuple(inputMax),
                                          partitioned=False)

            # load data onto MeshVariable
            if len(dset) == inputMesh.nodesGlobal:
                inputField = uw.mesh.MeshVariable( mesh=inputMesh, nodeDofCount=dof )
            elif  dset.shape[0] == inputMesh.subMesh.nodesGlobal:
                # load as a subMesh
                # assume the dset field belongs to the subMesh
                inputField = uw.mesh.MeshVariable( mesh=inputMesh.subMesh, nodeDofCount=dof )
            else:
                # raise error
                raise RuntimeError("The saved mesh file can't be read onto the interpolation grid.\n" \
                                   "Note: only subMesh variable with elementType 'DQ0' can be used presently used")

            # copy hdf5 numpy array onto serial inputField
            inputField.data[:] = dset[:]

            # interpolate 'inputField' onto the self nodes
            self.data[:] = inputField.evaluate(self.mesh.data)

        if units:
            if units.units == "degC":
                units = u.degK
                self.data[:] = nonDimensionalize((self.data[:]+273.15)*units)
            else:
                self.data[:] = nonDimensionalize(self.data[:]*units)

        uw.libUnderworld.StgFEM._FeVariable_SyncShadowValues( self._cself )
        h5f.close()

    def save(self, filename, meshHandle=None, units=None, time=None):
        """
        Save the MeshVariable to disk.

        Parameters
        ----------
        filename : string
            The name of the output file. Relative or absolute paths may be
            used, but all directories must exist.
        meshHandle :uw.utils.SavedFileData , optional
            The saved mesh file handle. If provided, a link is created within the
            mesh variable file to this saved mesh file. Important for checkpoint when
            the mesh deforms.

        Notes
        -----
        This method must be called collectively by all processes.

        Returns
        -------
        underworld.utils.SavedFileData
            Data object relating to saved file. This only needs to be retained
            if you wish to create XDMF files and can be ignored otherwise.

        Example
        -------
        First create the mesh add a variable:

        >>> mesh = uw.mesh.FeMesh_Cartesian( elementType='Q1/dQ0', elementRes=(16,16), minCoord=(0.,0.), maxCoord=(1.,1.) )
        >>> var = uw.mesh.MeshVariable( mesh=mesh, nodeDofCount=1, dataType="double" )

        Write something to variable

        >>> import numpy as np
        >>> var.data[:,0] = np.arange(var.data.shape[0])

        Save to a file (note that the 'ignoreMe' object isn't really required):

        >>> ignoreMe = var.save("saved_mesh_variable.h5")

        Now let's try and reload.

        >>> clone_var = uw.mesh.MeshVariable( mesh=mesh, nodeDofCount=1, dataType="double" )
        >>> clone_var.load("saved_mesh_variable.h5")

        Now check for equality:

        >>> np.allclose(var.data,clone_var.data)
        True

        >>> # clean up:
        >>> if uw.rank() == 0:
        ...     import os;
        ...     os.remove( "saved_mesh_variable.h5" )

        """
        if not isinstance(filename, str):
            raise TypeError("Expected 'filename' to be provided as a string")

        mesh = self.mesh
        h5f = h5py.File(name=filename, mode="w", driver='mpio', comm=MPI.COMM_WORLD)

        # ugly global shape def
        globalShape = ( mesh.nodesGlobal, self.data.shape[1] )
        # create dataset
        dset = h5f.create_dataset("data",
                                  shape=globalShape,
                                  dtype=self.data.dtype)
        fact = 1.0
        if units:
            fact = Dimensionalize(1.0, units=units).magnitude
            if units == "degC":
                fact = Dimensionalize(1.0, units=u.degK).magnitude
            # Save unit type as attribute
            h5f.attrs['units'] = str(units)

        if time:
            h5f.attrs['time'] = str(time)

        h5f.attrs["git commit"] = __git_revision__

        # write to the dset using the global node ids
        local = mesh.nodesLocal

        with dset.collective:
            if units == "degC":
                dset[mesh.data_nodegId[0:local],:] = self.data[0:local] * fact - 273.15
            else:
                dset[mesh.data_nodegId[0:local],:] = self.data[0:local] * fact

        # save a hdf5 attribute to the elementType used for this field - maybe useful
        h5f.attrs["elementType"] = np.string_(mesh.elementType)

        if hasattr( mesh.generator, "geometryMesh"):
            mesh = mesh.generator.geometryMesh

        if meshHandle:
            if not isinstance(meshHandle, (str, uw.utils.SavedFileData)):
                raise TypeError("Expected 'meshHandle' to be of type 'uw.utils.SavedFileData'")

            if isinstance(meshHandle, str):
                # DEPRECATION check
                import warnings
                warnings.warn("'meshHandle' paramater should be of type uw.utils.SaveFileData. Please update your models. "+
                              "Accepting 'meshHandle' as a string parameter will be removed in the next release.")
                meshFilename = meshHandle
            else:
                meshFilename = meshHandle.filename

            if not os.path.exists(meshFilename):
                raise ValueError("You are trying to link against the mesh file '{}'\n\
                                  that does not appear to exist. If you need to link \n\
                                  against a mesh file, please make sure it is created first.".format(meshFilename))
            # set reference to mesh (all procs must call following)
            h5f["mesh"] = h5py.ExternalLink(meshFilename, "./")

        h5f.close()

        # return our file handle
        return uw.utils.SavedFileData(self, filename)



from __future__ import print_function,  absolute_import


class BoundaryConditions(object):

    def __init__(self, Model, left=None, right=None, top=None, bottom=None,
                 front=None, back=None, nodeSets=tuple(),
                 order_wall_conditions=None):

        self.Model = Model
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom
        self.back = back
        self.front = front
        self.nodeSets = nodeSets

        if self.Model.mesh.dim == 2:
            self._wall_indexSets = {"bottom": (self.bottom,
                                               self.Model.bottom_wall),
                                    "top": (self.top,
                                            self.Model.top_wall),
                                    "left": (self.left,
                                             self.Model.left_wall),
                                    "right": (self.right,
                                              self.Model.right_wall)}
            if order_wall_conditions:
                if len(order_wall_conditions) <= 5:
                    self.order_wall_conditions = order_wall_conditions
            else:
                self.order_wall_conditions = ["bottom", "top", "left", "right"]

        if self.Model.mesh.dim == 3:
            self._wall_indexSets = {"bottom": (self.bottom,
                                               self.Model.bottom_wall),
                                    "top": (self.top,
                                            self.Model.top_wall),
                                    "left": (self.left,
                                             self.Model.left_wall),
                                    "right": (self.right,
                                              self.Model.right_wall),
                                    "front": (self.front,
                                              self.Model.front_wall),
                                    "back": (self.back,
                                             self.Model.back_wall)}
            if order_wall_conditions:
                if len(order_wall_conditions) <= 7:
                    self.order_wall_conditions = order_wall_conditions
            else:
                self.order_wall_conditions = ["bottom", "top", "front", "back",
                                              "left", "right"]



from __future__ import print_function,  absolute_import
import sys
from pyBadlands.model import Model as BadlandsModel
from scipy.interpolate import griddata, interp1d
from scipy.ndimage.filters import gaussian_filter
import numpy as np
from UWGeodynamics.scaling import COEFFICIENTS as scaling_coefficients
from UWGeodynamics.scaling import Dimensionalize, UnitRegistry

from tempfile import gettempdir
from .uw_utilities import get_UW_velocities

_tempdir = gettempdir()

from mpi4py import MPI

comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.rank

class SPM(object):

    def __init__(self, mesh, velocityField, swarm, materialField, airIndex,
                 sedimentIndex, XML, resolution, checkpoint_interval,
                 surfElevation=0., verbose=True, restartFolder=None,
                 restartStep=None):

        self.SECONDS_PER_YEAR = 31556925.9747  # Tropical year in seconds
        self.verbose = verbose
        self.restartStep = restartStep
        self.restartFolder = restartFolder

        # AutoScaling
        self.scaleDIM = 1.0 / scaling_coefficients["[length]"].magnitude
        self.scaleTIME = 1.0 / scaling_coefficients["[time]"].magnitude

        self.mesh = mesh
        self.velocityField = velocityField
        self.swarm = swarm
        self.material_index = materialField
        self.airIndex = airIndex
        self.sedimentIndex = sedimentIndex
        self.resolution = resolution
        self.surfElevation = surfElevation
        self.checkpoint_interval = checkpoint_interval / self.scaleTIME / self.SECONDS_PER_YEAR

        self.XML = XML

        if rank == 0:
            self.badlands_model = BadlandsModel()
            self.badlands_model.load_xml(self.XML)
            if self.restartStep:
                self.badlands_model.input.restart = True
                self.badlands_model.input.rstep = self.restartStep
                self.badlands_model.input.rfolder = self.restartFolder
                self.badlands_model.input.outDir = self.restartFolder
                self.badlands_model.outputStep = self.restartStep

        self.minCoord = self.mesh.minCoord
        self.maxCoord = self.mesh.maxCoord

        self.time_years = 0.
        if self.restartStep:
            # Parse xmf for the last timestep time
            import xml.etree.ElementTree as etree
            xmf = self.restartFolder+"/xmf/tin.time"+str(self.restartStep)+".xmf"
            tree = etree.parse(xmf)
            root = tree.getroot()
            self.time_years = float(root[0][0][0].attrib["Value"])

        self._tmp = _tempdir
        self._demfile = self._tmp+"/dem.csv"

        # Create Initial Flat DEM
        if rank == 0:
            self.dem = self._generate_flat_dem(self.minCoord,
                                               self.maxCoord,
                                               self.resolution,
                                               self.surfElevation,
                                               self.scaleDIM)

            np.savetxt(self._demfile, self.dem)
            # Build Mesh
            self.badlands_model.build_mesh(self._demfile, verbose=False)

            self.badlands_model.input.disp3d = True  # enable 3D displacements
            self.badlands_model.input.region = 0  # TODO: check what this does
            self.badlands_model.input.tStart = self.time_years
            self.badlands_model.tNow = self.time_years

            # Override the checkpoint/display interval in the Badlands model to
            # ensure BL and UW are synced
            self.badlands_model.input.tDisplay = self.checkpoint_interval

            # Set Badlands minimal distance between nodes before regridding
            self.badlands_model.force.merge3d = self.badlands_model.input.Afactor * self.badlands_model.recGrid.resEdges * 0.5

            # Bodge Badlands to perform an initial checkpoint
            # FIXME: we need to run the model for at least one iteration before this is generated. It would be nice if this wasn't the case.
            self.badlands_model.force.next_display = 0

        comm.Barrier()

        self._disp_inserted = False

        # Transfer the initial DEM state to Underworld
        self._update_material_types()

        comm.Barrier()

    def solve(self, dt, sigma=0):

        if rank == 0 and self.verbose:
            purple = "\033[0;35m"
            endcol = "\033[00m"
            print(purple+"Processing surface with Badlands"+endcol)
            sys.stdout.flush()

        dt_years = Dimensionalize(dt, UnitRegistry.years).magnitude

        if rank == 0:
            rg = self.badlands_model.recGrid
            if self.mesh.dim == 2:
                zVals = rg.regZ.mean(axis = 1)
                np_surface = np.column_stack((rg.regX, zVals)) * self.scaleDIM

            if self.mesh.dim == 3:
                np_surface = np.column_stack((rg.rectX, rg.rectY, rg.rectZ))*self.scaleDIM
        else:

            np_surface = None

        np_surface = comm.bcast(np_surface, root=0)
        comm.Barrier()

        # Get Velocity Field at the surface
        tracer_velocity_mps = get_UW_velocities(np_surface, self.velocityField) * self.scaleTIME / self.scaleDIM

        if rank == 0:
            # Use the tracer vertical velocities to deform the Badlands TIN
            # convert from meters per second to meters displacement over the whole iteration
            tracer_disp = tracer_velocity_mps * self.SECONDS_PER_YEAR * dt_years
            self._inject_badlands_displacement(self.time_years, dt_years, tracer_disp, sigma)

            # Run the Badlands model to the same time point
            self.badlands_model.run_to_time(self.time_years+dt_years)

        self.time_years += dt_years

        # TODO: Improve the performance of this function
        self._update_material_types()
        comm.Barrier()

        if rank == 0 and self.verbose:
            purple = "\033[0;35m"
            endcol = "\033[00m"
            print(purple + "Processing surface with Badlands...Done" + endcol)
            sys.stdout.flush()

        return

    def _determine_particle_state_2D(self):

        if rank == 0:
            known_xy = self.badlands_model.recGrid.tinMesh['vertices']*self.scaleDIM  # points that we have known elevation for
            known_z = self.badlands_model.elevation*self.scaleDIM  # elevation for those points
            xs = self.badlands_model.recGrid.regX*self.scaleDIM
            ys = self.badlands_model.recGrid.regY*self.scaleDIM
        else:
            known_xy = None
            known_z = None
            xs = None
            ys = None

        known_xy = comm.bcast(known_xy, root=0)
        known_z = comm.bcast(known_z, root=0)
        xs = comm.bcast(xs, root=0)
        ys = comm.bcast(ys, root=0)

        comm.Barrier()

        grid_x, grid_y = np.meshgrid(xs, ys)
        interpolate_z = griddata(known_xy, known_z, (grid_x, grid_y), method='nearest').T
        interpolate_z = interpolate_z.mean(axis=1)

        f = interp1d(xs, interpolate_z)

        uw_surface = self.swarm.particleCoordinates.data
        bdl_surface = f(uw_surface[:,0])

        flags = uw_surface[:,1] < bdl_surface

        return flags

    def _determine_particle_state(self):
        # Given Badlands' mesh, determine if each particle in 'volume' is above
        # (False) or below (True) it.

        # To do this, for each X/Y pair in 'volume', we interpolate its Z value
        # relative to the mesh in blModel. Then, if the interpolated Z is
        # greater than the supplied Z (i.e. Badlands mesh is above particle
        # elevation) it's sediment (True). Else, it's air (False).

        # TODO: we only support air/sediment layers right now; erodibility
        # layers are not implemented

        if rank == 0:
            known_xy = self.badlands_model.recGrid.tinMesh['vertices']*self.scaleDIM  # points that we have known elevation for
            known_z = self.badlands_model.elevation*self.scaleDIM  # elevation for those points
        else:
            known_xy = None
            known_z = None

        known_xy = comm.bcast(known_xy, root=0)
        known_z = comm.bcast(known_z, root=0)

        comm.Barrier()

        volume = self.swarm.particleCoordinates.data

        interpolate_xy = volume[:, [0, 1]]

        # NOTE: we're using nearest neighbour interpolation. This should be
        # sufficient as Badlands will normally run at a much higher resolution
        # than Underworld. 'linear' interpolation is much, much slower.
        interpolate_z = griddata(points=known_xy, values=known_z, xi=interpolate_xy, method='nearest')

        # True for sediment, False for air
        flags = volume[:, 2] < interpolate_z

        return flags

    def _update_material_types(self):

        # What do the materials (in air/sediment terms) look like now?
        if self.mesh.dim == 3:
            material_flags = self._determine_particle_state()
        if self.mesh.dim == 2:
            material_flags = self._determine_particle_state_2D()

        # If any materials changed state, update the Underworld material types
        mi = self.material_index.data

        # convert air to sediment
        for air_material in self.airIndex:
            sedimented_mask = np.logical_and(np.in1d(mi, air_material), material_flags)
            mi[sedimented_mask] = self.sedimentIndex

        # convert sediment to air
        for air_material in self.airIndex:
            eroded_mask = np.logical_and(~np.in1d(mi, air_material), ~material_flags)
            mi[eroded_mask] = self.airIndex[0]

    def _inject_badlands_displacement(self, time, dt, disp, sigma):
        """
        Takes a plane of tracer points and their DISPLACEMENTS in 3D over time
        period dt applies a gaussian filter on it. Injects it into Badlands as 3D
        tectonic movement.
        """

        # The Badlands 3D interpolation map is the displacement of each DEM
        # node at the end of the time period relative to its starting position.
        # If you start a new displacement file, it is treated as starting at
        # the DEM starting points (and interpolated onto the TIN as it was at
        # that tNow).


        # kludge; don't keep adding new entries
        if self._disp_inserted:
            self.badlands_model.force.T_disp[0, 0] = time
            self.badlands_model.force.T_disp[0, 1] = (time + dt)
        else:
            self.badlands_model.force.T_disp = np.vstack(([time, time + dt], self.badlands_model.force.T_disp))
            self._disp_inserted = True

        # Extent the velocity field in the third dimension
        if self.mesh.dim == 2:
            dispX = np.tile(disp[:,0], self.badlands_model.recGrid.rny)
            dispY = np.zeros((self.badlands_model.recGrid.rnx * self.badlands_model.recGrid.rny,))
            dispZ = np.tile(disp[:,1], self.badlands_model.recGrid.rny)

            disp = np.zeros((self.badlands_model.recGrid.rnx * self.badlands_model.recGrid.rny,3))
            disp[:,0] = dispX
            disp[:,1] = dispY
            disp[:,2] = dispZ

        # Gaussian smoothing
        if sigma>0:
            dispX = np.copy(disp[:,0]).reshape(self.badlands_model.recGrid.rnx, self.badlands_model.recGrid.rny)
            dispY = np.copy(disp[:,1]).reshape(self.badlands_model.recGrid.rnx, self.badlands_model.recGrid.rny)
            dispZ = np.copy(disp[:,2]).reshape(self.badlands_model.recGrid.rnx, self.badlands_model.recGrid.rny)
            smoothX = gaussian_filter(dispX, sigma)
            smoothY = gaussian_filter(dispY, sigma)
            smoothZ = gaussian_filter(dispZ, sigma)
            disp[:,0] = smoothX.flatten()
            disp[:,1] = smoothY.flatten()
            disp[:,2] = smoothZ.flatten()

        self.badlands_model.force.injected_disps = disp

    def _generate_flat_dem(self, minCoord, maxCoord, resolution, elevation, scale=1.):
        """
        Generate a flat DEM. This can be used as the initial Badlands state.

        minCoord: tuple of (X, Y, Z) coordinates defining the minimum bounds of
                  the DEM. Only the X and Y coordinates are used.
        maxCoord: tuple of (X, Y, Z) coordinates defining the maximum bounds of
                  the DEM. Only the X and Y coordinates are used.
        resolution: resolution of the model. Badlands assumes dx = dy.
        elevation: the Z parameter that each point is created at
        scale: the scaling factor between the 2 codes

        For your convenience, minCoord and maxCoord are designed to have the
        same formatting as the Underworld FeMesh_Cartesian minCoord and
        maxCoord parameters.

        IMPORTANT: minCoord and maxCoord are defined in terms of the Underworld
        coordinate system, but the returned DEM uses the Badlands coordinate
        system.

        Note that the initial elevation of the Badlands surface should coincide
        with the material transition in Underworld.
        """

        # Calculate number of nodes from required resolution.
        nx = np.int((maxCoord[0] - minCoord[0]) / resolution)
        ny = np.int((maxCoord[1] - minCoord[1]) / resolution)

        if self.mesh.dim == 2:
            minCoord = (minCoord[0], minCoord[0])
            maxCoord = (maxCoord[0], maxCoord[0])
            ny = nx

        items = []
        # FIXME: there should be a fast numpy way to do this
        for y in np.linspace(minCoord[1]/scale, maxCoord[1]/scale, ny):
            for x in np.linspace(minCoord[0]/scale, maxCoord[0]/scale, nx):
                items.append([x, y, elevation/scale])

        # NOTE: Badlands uses the difference in X coord of the first two points to determine the resolution.
        # This is something we should fix.
        # This is why we loop in y/x order instead of x/y order.
        return np.array(items)



from __future__ import print_function,  absolute_import
import underworld as uw
import numpy as np
from .scaling import nonDimensionalize as nd
from .scaling import UnitRegistry as u
from ._boundary_conditions import BoundaryConditions

_dim_temp = {"[temperature]":1.0}
_dim_heat_flux = {"[mass]":1.0, "[time]":-3.0}

class TemperatureBCs(BoundaryConditions):

    def __init__(self, Model, left=None, right=None, top=None, bottom=None,
                 front=None, back=None, nodeSets=None, materials=None,
                 bottom_material=None, top_material=None, left_material=None,
                 right_material=None, back_material=None, front_material=None,
                 order_wall_conditions=None):

        super(TemperatureBCs, self).__init__(Model, left, right, top, bottom, front, back,
                                             nodeSets, order_wall_conditions)

        self.materials = materials
        self.bottom_material = bottom_material
        self.top_material = top_material
        self.left_material = left_material
        self.right_material = right_material
        self.back_material = back_material
        self.front_material = front_material

        self._material_boundaries = {"left": left_material,
                                     "right": right_material,
                                     "top": top_material,
                                     "bottom": bottom_material,
                                     "back": back_material,
                                     "front": front_material}

        for key in self._material_boundaries.keys():
            wall = self.__dict__[key]
            if wall:
                wall_material = self._material_boundaries[key]
                if (wall.dimensionality == _dim_heat_flux) and not(wall and wall_material):
                    raise ValueError(
                        """left wall is {0} and left_material is {1}""".format(
                            wall, wall_material))

    def __getitem__(self, name):
        return self.__dict__[name]

    @staticmethod
    def _get_heat_flux(heat_flow, material):

        if not material.capacity:
            raise ValueError("""Material {0} has no capacity
                             defined""".format(material.name))

        if not material.density.reference_density:
            raise ValueError("""Material {0} has no density
                             defined""".format(material.name))

        cp = material.capacity
        rho = material.density.reference_density
        return heat_flow / (rho * cp)

    def get_conditions(self):

        Model = self.Model
        # Reinitialise neumnann and dirichlet condition
        self.dirichlet_indices = []
        self.neumann_indices = []

        self.dirichlet_indices.append(Model.mesh.specialSets["Empty"])
        self.neumann_indices.append(Model.mesh.specialSets["Empty"])

        for wall in self.order_wall_conditions:
            condition, indexSet = self._wall_indexSets[wall]
            if condition:
                if condition.dimensionality == _dim_temp:
                    Model.temperature.data[indexSet.data] = nd(condition)
                    self.dirichlet_indices[0] += indexSet
                elif condition.dimensionality == _dim_heat_flux:
                    material = self._material_boundaries[wall]
                    self.neumann_indices[0] += indexSet
                    heat_flow = self._get_heat_flux(condition, material)
                    Model._heatFlux.data[indexSet.data] = nd(heat_flow)

        if self.nodeSets:
            for elem in self.nodeSets:
                nodeSet, temp = elem
                isThere = np.in1d(Model.mesh.data_nodegId, list(nodeSet))
                if isThere.any():
                    local_indices = np.arange(Model.mesh.nodesDomain)[isThere]
                    indexSet = uw.mesh.FeMesh_IndexSet(
                        self.Model.mesh, topologicalIndex=0,
                        size=self.Model.mesh.nodesGlobal,
                        fromObject=local_indices)
                    Model.temperature.data[local_indices] = nd(temp)
                    self.dirichlet_indices[0] += indexSet

        if self.materials:
            for (material, temp) in self.materials:
                if material and nd(temp):
                    indexSet = Model._get_material_indices(material)
                    Model.temperature.data[indexSet.data] = nd(temp)
                    self.dirichlet_indices[0] += indexSet

        conditions = []

        conditions.append(uw.conditions.DirichletCondition(
            variable=Model.temperature,
            indexSetsPerDof=self.dirichlet_indices))

        neumann_indices = []
        for val in self.neumann_indices:
            if val.data.size > 0:
                neumann_indices.append(val)
            else:
                neumann_indices.append(None)
        neumann_indices = tuple(neumann_indices)

        if neumann_indices != tuple([None]):
            conditions.append(uw.conditions.NeumannCondition(
                fn_flux=Model._heatFlux,
                variable=Model.temperature,
                indexSetsPerDof=self.neumann_indices))

        if not conditions:
            raise ValueError("Undefined conditions")

        return conditions




from __future__ import print_function,  absolute_import
import underworld as uw
import h5py
from mpi4py import MPI
from UWGeodynamics.scaling import Dimensionalize
from UWGeodynamics.scaling import nonDimensionalize
from UWGeodynamics.scaling import UnitRegistry as u
from UWGeodynamics.version import git_revision as __git_revision__
from . import _meshvariable as var

class FeMesh_Cartesian(uw.mesh.FeMesh_Cartesian):

    def __init__(self, elementType="Q1/dQ0",
                 elementRes=(4,4), minCoord=(0.,0.),
                 maxCoord=(1.,1.), periodic=None,
                 partitioned=True, **kwargs):

        super(FeMesh_Cartesian, self).__init__(elementType,
                                               elementRes,
                                               minCoord,
                                               maxCoord,
                                               periodic,
                                               partitioned,
                                               **kwargs)

    def add_variable(self, nodeDofCount, dataType='double', **kwargs):
        """
        Creates and returns a mesh variable using the discretisation of the given mesh.

        To set / read nodal values, use the numpy interface via the 'data' property.

        Parameters
        ----------
        dataType : string
            The data type for the variable.
            Note that only 'double' type variables are currently
            supported.
        nodeDofCount : int
            Number of degrees of freedom per node the variable will have

        Returns
        -------
        underworld.mesh.MeshVariable
            The newly created mesh variable.

        Example
        -------
        >>> linearMesh  = uw.mesh.FeMesh_Cartesian( elementType='Q1/dQ0', elementRes=(16,16), minCoord=(0.,0.), maxCoord=(1.,1.) )
        >>> scalarFeVar = linearMesh.add_variable( nodeDofCount=1, dataType="double" )
        >>> q0field     = linearMesh.subMesh.add_variable( 1 )  # adds variable to secondary elementType discretisation
        """

        return  var.MeshVariable(self, nodeDofCount, dataType, **kwargs)

    def save(self, filename, units=None, time=None):
        """
        Save the mesh to disk

        Parameters
        ----------
        filename : string
            The name of the output file.

        Returns
        -------
        underworld.utils.SavedFileData
            Data object relating to saved file. This only needs to be retained
            if you wish to create XDMF files and can be ignored otherwise.

        Notes
        -----
        This method must be called collectively by all processes.

        Example
        -------
        First create the mesh:

        >>> mesh = uw.mesh.FeMesh_Cartesian( elementType='Q1/dQ0', elementRes=(16,16), minCoord=(0.,0.), maxCoord=(1.,1.) )

        Save to a file (note that the 'ignoreMe' object isn't really required):

        >>> ignoreMe = mesh.save("saved_mesh.h5")

        Now let's try and reload. First create new mesh (note the different spatial size):

        >>> clone_mesh = uw.mesh.FeMesh_Cartesian( elementType='Q1/dQ0', elementRes=(16,16), minCoord=(0.,0.), maxCoord=(1.5,1.5) )

        Confirm clone mesh is different from original mesh:

        >>> import numpy as np
        >>> np.allclose(mesh.data,clone_mesh.data)
        False

        Now reload using saved file:

        >>> clone_mesh.load("saved_mesh.h5")

        Now check for equality:

        >>> np.allclose(mesh.data,clone_mesh.data)
        True

        >>> # clean up:
        >>> if uw.rank() == 0:
        ...     import os;
        ...     os.remove( "saved_mesh.h5" )

        """

        if hasattr(self.generator, 'geometryMesh'):
            raise RuntimeError("Cannot save this mesh as it's a subMesh. "
                                + "Most likely you only need to save its geometryMesh")
        if not isinstance(filename, str):
            raise TypeError("'filename', must be of type 'str'")

        h5f = h5py.File(name=filename, mode="w", driver='mpio', comm=MPI.COMM_WORLD)

        fact = 1.0
        if units:
            fact = Dimensionalize(1.0, units=units).magnitude
            h5f.attrs['units'] = str(units)

        # save attributes and simple data - MUST be parallel as driver is mpio
        h5f.attrs['dimensions'] = self.dim
        h5f.attrs['mesh resolution'] = self.elementRes
        h5f.attrs['max'] = tuple([fact*x for x in self.maxCoord])
        h5f.attrs['min'] = tuple([fact*x for x in self.minCoord])
        h5f.attrs['regular'] = self._cself.isRegular
        h5f.attrs['elementType'] = self.elementType
        h5f.attrs['time'] = str(time)
        h5f.attrs["git commit"] = __git_revision__

        # write the vertices
        globalShape = ( self.nodesGlobal, self.data.shape[1] )
        dset = h5f.create_dataset("vertices",
                                  shape=globalShape,
                                  dtype=self.data.dtype)

        local = self.nodesLocal
        # write to the dset using the local set of global node ids
        with dset.collective:
            dset[self.data_nodegId[0:local],:] = self.data[0:local] * fact

        # write the element node connectivity
        globalShape = ( self.elementsGlobal, self.data_elementNodes.shape[1] )
        dset = h5f.create_dataset("en_map",
                                  shape=globalShape,
                                  dtype=self.data_elementNodes.dtype)

        local = self.elementsLocal
        # write to the dset using the local set of global node ids
        with dset.collective:
            dset[self.data_elgId[0:local], :] = self.data_elementNodes[0:local]

        h5f.close()

        # return our file handle
        return uw.utils.SavedFileData(self, filename)

    def load(self, filename):
        """
        Load the mesh from disk.

        Parameters
        ----------
        filename: str
            The filename for the saved file. Relative or absolute paths may be
            used, but all directories must exist.

        Notes
        -----
        This method must be called collectively by all processes.

        If the file data array is the same length as the current mesh
        global size, it is assumed the file contains compatible data. Note that
        this may not be the case where for example where you have saved using a
        2*8 resolution mesh, then loaded using an 8*2 resolution mesh.

        Provided files must be in hdf5 format, and use the correct schema.

        Example
        -------
        Refer to example provided for 'save' method.

        """
        self.reset()
        if not isinstance(filename, str):
            raise TypeError("Expected filename to be provided as a string")

        # get field and mesh information
        h5f = h5py.File( filename, "r", driver='mpio', comm=MPI.COMM_WORLD );

        # get resolution of old mesh
        res = h5f.attrs['mesh resolution']
        if res is None:
            raise RuntimeError("Can't read the 'mesh resolution' for the field hdf5 file,"+
                   " was it created correctly?")

        if (res == self.elementRes).all() == False:
            raise RuntimeError("Provided file mesh resolution does not appear to correspond to\n"\
                               "resolution of mesh object.")

        # get units
        try:
            units = h5f.attrs["units"]
        except KeyError:
            units = None

        if units and units != "None":
            units = u.parse_expression(units)
        else:
            units = None

        dset = h5f.get('vertices')
        if dset == None:
            raise RuntimeError("Can't find the 'vertices' dataset in hdf5 file '{0}'".format(filename) )

        dof = dset.shape[1]
        if dof != self.data.shape[1]:
            raise RuntimeError("Can't load hdf5 '{0}', incompatible data shape".format(filename))

        if len(dset) != self.nodesGlobal:
            raise RuntimeError("Provided data file appears to be for a different resolution mesh.")

        with self.deform_mesh(isRegular=h5f.attrs['regular']):
            with dset.collective:
                if units:
                    self.data[0:self.nodesLocal] = nonDimensionalize(
                        dset[self.data_nodegId[0:self.nodesLocal], :] * units)
                else:
                    self.data[0:self.nodesLocal] = dset[self.data_nodegId[0:self.nodesLocal], :]

        h5f.close()



from __future__ import print_function,  absolute_import
import underworld.function as fn
from .scaling import nonDimensionalize as nd
from .scaling import UnitRegistry as u


class Density(object):

    def __init__(self, temperatureField=None, pressureField=None):
        self.temperatureField = temperatureField
        self.pressureField = pressureField
        self.name = None


class ConstantDensity(Density):

    def __init__(self, reference_density):
        self.reference_density = reference_density
        self._density = nd(reference_density)
        self.name = "Constant ({0})".format(str(reference_density))

    def effective_density(self):
        return fn.misc.constant(self._density)


class LinearDensity(Density):

    def __init__(self, reference_density, thermalExpansivity=3e-5 / u.kelvin,
                 reference_temperature=273.15 * u.degK, beta=0. / u.pascal,
                 reference_pressure=0. * u.pascal):

        """ The LinearDensity function calculates:
            density = rho0 * (1 + (beta * deltaP) - (alpha * deltaT))
            where deltaP is the difference between P and the reference P, and deltaT
            is the difference between T and the reference T."""

        super(LinearDensity, self).__init__()

        self.name = "Linear (ref: {0})".format(str(reference_density))
        self.reference_density = reference_density
        self.reference_temperature = reference_temperature
        self.thermalExpansivity = thermalExpansivity
        self.reference_pressure = reference_pressure
        self._alpha = nd(thermalExpansivity)
        self._beta = nd(beta)
        self._Tref = nd(reference_temperature)
        self._Pref = nd(reference_pressure)

    def effective_density(self):

        density = nd(self.reference_density)

        # Temperature dependency
        if not self.temperatureField:
            raise RuntimeError("No temperatureField found!")

        t_term = self._alpha * (self.temperatureField - self._Tref)

        # Pressure dependency
        if not self.pressureField:
            raise RuntimeError("No pressureField found!")

        p_term = self._beta * (self.pressureField - self._Pref)

        return density * (1.0 + p_term - t_term)




from __future__ import print_function,  absolute_import
import numpy as np
import underworld as uw
import underworld.function as fn

class FrictionBoundaries(object):
    """ This class flags elements at the boundaries

    """
    def __init__(self, Model, rightFriction=None, leftFriction=None,
                 topFriction=None, bottomFriction=None, frontFriction=None,
                 backFriction=None, thickness=2):

        self.Model = Model
        self.thickness = thickness

        # Build borders
        globalIndices = np.arange(np.prod(Model.mesh.elementRes))
        globalIndices = globalIndices.reshape((Model.mesh.elementRes[::-1]))

        self.bottomFriction = bottomFriction
        self.rightFriction = rightFriction
        self.leftFriction = leftFriction
        self.topFriction = topFriction
        self.frontFriction = frontFriction
        self.backFriction = backFriction

        self.subMesh = Model.mesh.subMesh
        conditions = list()

        self._mask = uw.mesh.MeshVariable(mesh=self.subMesh, nodeDofCount=1)
        self._mask.data[:] = 0

        if self.rightFriction:
            self._right_mask = uw.mesh.MeshVariable(mesh=self.subMesh, nodeDofCount=1)
            self._right_mask.data[:] = 0

            if Model.mesh.dim < 3:
                right = globalIndices[:, -thickness:].ravel()
            else:
                right = globalIndices[:, :, -thickness:].ravel()

            intersect = np.intersect1d(self.subMesh.data_nodegId.ravel(), right)
            mask = np.in1d(self.subMesh.data_nodegId.ravel(), intersect)
            self._right_mask.data[mask, 0] = 1
            self._mask.data[mask, 0] = 1
            conditions.append((self._right_mask > 0., self.rightFriction))

        if self.leftFriction:
            self._left_mask = uw.mesh.MeshVariable(mesh=self.subMesh, nodeDofCount=1)
            self._left_mask.data[:] = 0

            if Model.mesh.dim < 3:
                left = globalIndices[:, :thickness].ravel()
            else:
                left = globalIndices[:, :, :thickness].ravel()

            intersect = np.intersect1d(self.subMesh.data_nodegId.ravel(), left)
            mask = np.in1d(self.subMesh.data_nodegId.ravel(), intersect)
            self._left_mask.data[mask, 0] = 1
            self._mask.data[mask, 0] = 1
            conditions.append((self._left_mask > 0., self.leftFriction))

        if self.frontFriction:
            self._front_mask = uw.mesh.MeshVariable(mesh=self.subMesh, nodeDofCount=1)
            self._front_mask.data[:] = 0

            if Model.mesh.dim > 2:
                front = globalIndices[:, :thickness, :].ravel()
            else:
                raise ValueError("Mesh is 2D")

            intersect = np.intersect1d(self.subMesh.data_nodegId.ravel(), front)
            mask = np.in1d(self.subMesh.data_nodegId.ravel(), intersect)
            self._front_mask.data[mask, 0] = 1
            self._mask.data[mask, 0] = 1
            conditions.append((self._front_mask > 0., self.frontFriction))

        if self.backFriction:
            self._back_mask = uw.mesh.MeshVariable(mesh=self.subMesh, nodeDofCount=1)
            self._back_mask.data[:] = 0

            if Model.mesh.dim > 2:
                back = globalIndices[:, -thickness:, :].ravel()
            else:
                raise ValueError("Mesh is 2D")

            intersect = np.intersect1d(self.subMesh.data_nodegId.ravel(), back)
            mask = np.in1d(self.subMesh.data_nodegId.ravel(), intersect)
            self._back_mask.data[mask, 0] = 1
            self._mask.data[mask, 0] = 1
            conditions.append((self._back_mask > 0., self.backFriction))

        if self.bottomFriction:
            self._bottom_mask = uw.mesh.MeshVariable(mesh=self.subMesh, nodeDofCount=1)
            self._bottom_mask.data[:] = 0

            bottom = globalIndices[:thickness].ravel()
            # Take the intersection between the globalID and the boundaries where
            # friction is to be applied
            intersect = np.intersect1d(self.subMesh.data_nodegId.ravel(), bottom)
            # Create a mask to highlight those elements in the local domain
            mask = np.in1d(self.subMesh.data_nodegId.ravel(), intersect)
            self._bottom_mask.data[mask, 0] = 1
            self._mask.data[mask, 0] = 1
            conditions.append((self._bottom_mask > 0., self.bottomFriction))

        if self.topFriction:
            self._top_mask = uw.mesh.MeshVariable(mesh=self.subMesh, nodeDofCount=1)
            self._top_mask.data[:] = 0

            top = globalIndices[-thickness:].ravel()
            intersect = np.intersect1d(self.subMesh.data_nodegId.ravel(), top)
            mask = np.in1d(self.subMesh.data_nodegId.ravel(), intersect)
            self._top_mask.data[mask, 0] = 1
            self._mask.data[mask, 0] = 1
            conditions.append((self._top_mask > 0., self.topFriction))

        conditions.append((True, -1.0))
        self.friction = fn.branching.conditional(conditions)




# THIS FILE IS GENERATED FROM UWGeodynamics SETUP.PY
#
short_version = '0.9.8b'
version = '0.9.8b'
full_version = '0.9.8b-dev-41bb885(master)'
git_revision = '41bb8852d31c1c2e6a27d684217bd2ade9f82e19'
release = False

if not release:
    version = full_version



from __future__ import print_function, absolute_import
import os
import sys
from collections import OrderedDict
import numpy as np
import underworld as uw
import underworld.function as fn
from underworld.function.exception import SafeMaths as Safe
import UWGeodynamics.shapes as shapes
import UWGeodynamics.surfaceProcesses as surfaceProcesses
from . import rcParams
from .scaling import Dimensionalize
from .scaling import nonDimensionalize as nd
from .scaling import UnitRegistry as u
from .lithopress import LithostaticPressure
from ._utils import PressureSmoother, PassiveTracers, PassiveTracersGrid
from ._rheology import ViscosityLimiter, StressLimiter
from ._material import Material
from ._visugrid import Visugrid
from ._velocity_boundaries import VelocityBCs
from ._velocity_boundaries import StressBCs
from ._thermal_boundaries import TemperatureBCs
from ._mesh_advector import _mesh_advector
from ._frictional_boundary import FrictionBoundaries
from .Underworld_extended import FeMesh_Cartesian
from .Underworld_extended import Swarm
from .Underworld_extended import MeshVariable
from .Underworld_extended import SwarmVariable
from datetime import datetime
from .version import full_version
from ._freesurface import FreeSurfaceProcessor
from mpi4py import MPI

_dim_gravity = {'[length]': 1.0, '[time]': -2.0}
_dim_time = {'[time]': 1.0}


class Model(Material):
    """UWGeodynamic Model Class"""

    @u.check([None, (None, None), ("[length]", "[length]"), None,
              _dim_gravity])
    def __init__(self, elementRes=(64, 64),
                 minCoord=(0., 0.), maxCoord=(64. * u.kilometer, 64 * u.kilometer),
                 name=None, gravity=None, periodic=None, elementType=None,
                 temperatureBCs=None, velocityBCs=None, stressBCs=None, materials=None,
                 outputDir=None, frictionalBCs=None, surfaceProcesses=None,
                 isostasy=None, visugrid=None):
        """__init__

        Parameters
        ----------

        elementRes :
            Resolution of the mesh in number of elements for each axis (degree
            of freedom)
        minCoord :
            Minimum coordinates for each axis.
        maxCoord :
            Maximum coordinates for each axis.
        name :
            The Model name.
        gravity :
            Acceleration due to gravity vector.
        periodic :
            Mesh periodicity.
        elementType :
            Type of finite element.
        temperatureBCs :
            Temperature Boundary Condition, must be object of type:
            TemperatureBCs
        velocityBCs :
            Velocity Boundary Condtion, must be object of type VelocityBCs
        stressBCs :
            Stress Boundary Condtion, must be object of type StressBCs
        materials :
            List of materials, each material must be an object of type:
            Material
        outputDir :
            Output Directory
        frictionalBCs :
            Frictional Boundary Conditions, must be object of type:
            FrictionBoundaries
        surfaceProcesses :
            Surface Processes, must be an object of type:
            surfaceProcesses
        isostasy :
            Isostasy Solver
        visugrid :
            Visugrid object
        advector :
            Mesh advector object

        Examples
        --------
        >>> import UWGeodynamics as GEO
        >>> u = GEO.UnitRegistry
        >>> Model = Model = GEO.Model(
                elementRes=(64, 64),
                minCoord=(0., 0.),
                maxCoord=(64. * u.kilometer, 64. * u.kilometer))

        """

        super(Model, self).__init__()

        # Process __init__ arguments
        if not name:
            self.name = rcParams["model.name"]
        else:
            self.name = name

        self.minCoord = minCoord
        self.maxCoord = maxCoord
        self.top = maxCoord[-1]
        self.bottom = minCoord[-1]
        self.dimension = dim = len(maxCoord)

        if not gravity:
            self.gravity = [0.0 for val in range(self.dimension)]
            self.gravity[-1] = -1.0 * rcParams["gravity"]
            self.gravity = tuple(self.gravity)
        else:
            self.gravity = gravity

        if not elementType:
            self.elementType = rcParams["element.type"]
        else:
            self.elementType = elementType

        self.elementRes = elementRes

        if outputDir:
            self.outputDir = outputDir
        else:
            self.outputDir = rcParams["output.directory"]

        # Compute model dimensions
        self.length = maxCoord[0] - minCoord[0]
        self.height = maxCoord[-1] - minCoord[-1]

        if dim == 3:
            self.width = maxCoord[1] - minCoord[1]

        if periodic:
            self.periodic = periodic
        else:
            periodic = tuple([False for val in range(dim)])
            self.periodic = periodic

        # Get non-dimensional extents along each axis
        minCoord = tuple([nd(val) for val in self.minCoord])
        maxCoord = tuple([nd(val) for val in self.maxCoord])

        # Initialize model mesh
        self.mesh = FeMesh_Cartesian(elementType=self.elementType,
                                     elementRes=self.elementRes,
                                     minCoord=minCoord,
                                     maxCoord=maxCoord,
                                     periodic=self.periodic)

        self.mesh_fields = {}
        self.submesh_fields = {}
        self.swarm_fields = {}

        # Add common mesh variables
        self.temperature = False
        self.add_submesh_field("pressureField", nodeDofCount=1)
        self.add_mesh_field("velocityField", nodeDofCount=self.mesh.dim)
        self.add_mesh_field("tractionField", nodeDofCount=self.mesh.dim)
        self.add_submesh_field("_strainRateField", nodeDofCount=1)

        # Create a field to check applied boundary conditions
        self.add_mesh_field("boundariesField", nodeDofCount=self.mesh.dim)

        # symmetric component of the gradient of the flow velocityField.
        self.strainRate = fn.tensor.symmetric(self.velocityField.fn_gradient)
        self._strainRate_2ndInvariant = None

        # Create the material swarm
        self.swarm = Swarm(mesh=self.mesh, particleEscape=True)
        if self.mesh.dim == 2:
            particlesPerCell = rcParams["swarm.particles.per.cell.2D"]
        else:
            particlesPerCell = rcParams["swarm.particles.per.cell.3D"]

        self._swarmLayout = uw.swarm.layouts.PerCellSpaceFillerLayout(
            swarm=self.swarm,
            particlesPerCell=particlesPerCell)

        self.swarm.populate_using_layout(layout=self._swarmLayout)

        # timing and checkpointing
        self.checkpointID = 0
        self.time = 0.0 * u.megayears
        self.step = 0
        self._dt = None

        self._defaultMaterial = self.index
        self.materials = list(materials) if materials is not None else list()
        self.materials.append(self)

        # Create a series of aliases for the boundary sets
        self.left_wall = self.mesh.specialSets["MinI_VertexSet"]
        self.right_wall = self.mesh.specialSets["MaxI_VertexSet"]

        if self.mesh.dim == 2:
            self.top_wall = self.mesh.specialSets["MaxJ_VertexSet"]
            self.bottom_wall = self.mesh.specialSets["MinJ_VertexSet"]
        else:
            self.front_wall = self.mesh.specialSets["MinJ_VertexSet"]
            self.back_wall = self.mesh.specialSets["MaxJ_VertexSet"]
            self.top_wall = self.mesh.specialSets["MaxK_VertexSet"]
            self.bottom_wall = self.mesh.specialSets["MinK_VertexSet"]

        # Boundary Conditions
        self.velocityBCs = velocityBCs
        self.stressBCs = stressBCs
        self.temperatureBCs = temperatureBCs
        self.frictionalBCs = frictionalBCs
        self._isostasy = isostasy
        self.surfaceProcesses = surfaceProcesses

        self.pressSmoother = PressureSmoother(self.mesh, self.pressureField)

        # Passive Tracers
        # An ordered dict is required to ensure that all threads process
        # the same swarm at a time.
        self.passive_tracers = OrderedDict()

        # Visualisation
        self._visugrid = visugrid

        # Mesh advector
        self._advector = None

        # init solver
        self._solver = None
        self._static_solver = False

        # Initialise remaining attributes
        self.defaultStrainRate = 1e-15 / u.second
        self._solution_exist = fn.misc.constant(False)
        self._isYielding = None
        self._temperatureDot = None
        self._temperature = None
        self.DiffusivityFn = None
        self.HeatProdFn = None
        self._buoyancyFn = None
        self._freeSurface = False
        self.callback_post_solve = None
        self._mesh_saved = False
        self._initialize()

    def _initialize(self):
        """_initialize
        Model Initialisation
        """

        self.swarm_advector = uw.systems.SwarmAdvector(
            swarm=self.swarm,
            velocityField=self.velocityField,
            order=2
        )

        if self.mesh.dim == 2:
            particlesPerCell = rcParams["popcontrol.particles.per.cell.2D"]
        else:
            particlesPerCell = rcParams["popcontrol.particles.per.cell.3D"]

        self.population_control = uw.swarm.PopulationControl(
            self.swarm,
            aggressive=rcParams["popcontrol.aggressive"],
            splitThreshold=rcParams["popcontrol.split.threshold"],
            maxSplits=rcParams["popcontrol.max.splits"],
            particlesPerCell=particlesPerCell)

        # Add Common Swarm Variables
        self.add_swarm_field("materialField", dataType="int", count=1,
                             init_value=self.index)
        self.add_swarm_field("plasticStrain", dataType="double", count=1)
        self.add_swarm_field("_viscosityField", dataType="double", count=1)
        self.add_swarm_field("_densityField", dataType="double", count=1)
        self.add_swarm_field("meltField", dataType="double", count=1)
        self.add_swarm_field("timeField", dataType="double", count=1)
        self.timeField.data[...] = 0.0
        self.materialField.data[...] = self.index

        if self.mesh.dim == 3:
            stress_dim = 6
        else:
            stress_dim = 3

        self.add_swarm_field("_previousStressField", dataType="double",
                             count=stress_dim)
        self.add_swarm_field("_stressTensor", dataType="double",
                             count=stress_dim, projected="submesh")
        self.add_swarm_field("_stressField", dataType="double",
                             count=1, projected="submesh")

    def __getitem__(self, name):
        """__getitem__

        Return item with name=name from the class __dict__
        This allows the user to get the attributes of the model
        class as:
            Model["name"]

        Parameters
        ----------

        name : name of the attribute

        Returns
        -------
        Attribute of the Model instance.
        self.__dict__[name]
        """
        return self.__dict__[name]

    def _repr_html_(self):
        """_repr_html_

        HTML table describing the model.
        For integration with Jupyter notebook.
        """
        return _model_html_repr(self)

    @property
    def x(self):
        """x"""
        return fn.input()[0]

    @property
    def y(self):
        """y"""
        return fn.input()[1]

    @property
    def z(self):
        """z"""
        return fn.input()[2]

    @property
    def outputDir(self):
        """ Output Directory """
        return self._outputDir

    @outputDir.setter
    def outputDir(self, value):
        self._outputDir = value

    def restart(self, step, restartDir):
        """restart

        Parameters
        ----------

        step :
                Step from which you want to restart the model.
                Must be an int (step number either absolute or relative)
                if step == -1, run the last available step
                if step == -2, run the second last etc.
        restartDir :
                Directory that contains the outputs of the model
                you want to restart from.

        Returns
        -------

        This function returns None
        """

        if not isinstance(step, int):
            raise ValueError("step must be an int or a bool")

        if not os.path.exists(restartDir):
            raise ValueError("restartDir must be a path to an existing folder")

        # Do not raise and error idf directory is empty, just run
        # the model...
        if not os.listdir(restartDir):
            return

        # Look for step with swarm available
        indices = [int(os.path.splitext(filename)[0].split("-")[-1])
                   for filename in os.listdir(restartDir) if "-" in
                   filename]
        indices.sort()

        if step < 0:
            step = indices[step]

        if step not in indices:
            raise ValueError("Cannot find step in specified folder")

        # Ready for restart

        # Get time from swarm-%.h5 file
        import h5py
        with h5py.File(os.path.join(restartDir, "swarm-%s.h5" % step), "r",
                       driver="mpio", comm=MPI.COMM_WORLD) as h5f:
            self.time = u.Quantity(h5f.attrs.get("time"))

        if uw.rank() == 0:
            print(80 * "=" + "\n")
            print("Restarting Model from Step {0} at Time = {1}\n".format(step,self.time))
            print('(' + datetime.now().strftime('%Y-%m-%d %H:%M:%S') + ')')
            print(80 * "=" + "\n")
            sys.stdout.flush()

        self.checkpointID = step

        # Reload deformed mesh if advector is present
        if self._advector:
            self.mesh.load(os.path.join(restartDir, 'mesh-%s.h5' % step))
        # Load mesh
        else:
            self.mesh.load(os.path.join(restartDir, "mesh.h5"))

        if uw.rank() == 0:
            print("Mesh loaded" + '(' + datetime.now().strftime('%Y-%m-%d %H:%M:%S') + ')')
            sys.stdout.flush()

        self.swarm = Swarm(mesh=self.mesh, particleEscape=True)
        self.swarm.load(os.path.join(restartDir, 'swarm-%s.h5' % step))

        if uw.rank() == 0:
            print("Swarm loaded" + '(' + datetime.now().strftime('%Y-%m-%d %H:%M:%S') + ')')
            sys.stdout.flush()

        self._initialize()

        # Reload all the restart fields
        for field in rcParams["restart.fields"]:
            if field == "temperature":
                continue
            obj = getattr(self, field)
            path = os.path.join(restartDir, field + "-%s.h5" % step)
            if uw.rank() == 0:
                print("Reloading field {0} from {1}".format(field, path))
                sys.stdout.flush()
            obj.load(str(path))
            if uw.rank() == 0:
                print("{0} loaded".format(field) + '(' + datetime.now().strftime('%Y-%m-%d %H:%M:%S') + ')')
                sys.stdout.flush()

        # Temperature is a special case...
        path = os.path.join(restartDir, "temperature" + "-%s.h5" % step)
        if os.path.exists(path):
            if not self.temperature:
                self.temperature = True
            obj = getattr(self, "temperature")
            if uw.rank() == 0:
                print("Reloading field {0} from {1}".format("temperature", path))
                sys.stdout.flush()
            obj.load(str(path))
            if uw.rank() == 0:
                print("Temperature loaded" + '(' +
                        datetime.now().strftime('%Y-%m-%d %H:%M:%S') + ')')
                sys.stdout.flush()

        # Reload Passive Tracers
        for key, tracer in self.passive_tracers.items():

            if uw.rank() == 0:
                print("Reloading {0} passive tracers".format(tracer.name))
                sys.stdout.flush()

            fname = tracer.name + '-%s.h5' % step
            fpath = os.path.join(restartDir, fname)
            with h5py.File(fpath, "r", driver="mpio", comm=MPI.COMM_WORLD) as h5f:
                vertices = h5f["data"].value * u.Quantity(h5f.attrs["units"])
                vertices = [vertices[:, dim] for dim in range(self.mesh.dim)]
                obj = PassiveTracers(self.mesh,
                                     self.velocityField,
                                     tracer.name,
                                     vertices=vertices,
                                     particleEscape=tracer.particleEscape)

            attr_name = tracer.name.lower() + "_tracers"
            setattr(self, attr_name, obj)
            self.passive_tracers[key] = obj
            if uw.rank() == 0:
                print("{0} loaded".format(tracer.name) + '(' + datetime.now().strftime('%Y-%m-%d %H:%M:%S') + ')')
                sys.stdout.flush()

        if isinstance(self.surfaceProcesses,
                      (surfaceProcesses.SedimentationThreshold,
                       surfaceProcesses.ErosionThreshold,
                       surfaceProcesses.ErosionAndSedimentationThreshold)):

            obj = self.surfaceProcesses
            obj.Model = self
            obj.timeField = self.timeField

        # Restart Badlands if we are running a coupled model
        if isinstance(self.surfaceProcesses, surfaceProcesses.Badlands):
            badlands_model = self.surfaceProcesses
            restartFolder = badlands_model.restartFolder
            restartStep = badlands_model.restartStep

            # Parse xmf for the last timestep time
            import xml.etree.ElementTree as etree
            xmf = restartFolder + "/xmf/tin.time" + str(restartStep) + ".xmf"
            tree = etree.parse(xmf)
            root = tree.getroot()
            badlands_time = float(root[0][0][0].attrib["Value"])
            uw_time = self.time.to(u.years).magnitude

            if np.abs(badlands_time - uw_time) > 1:
                raise ValueError("""Time in Underworld and Badlands outputs
                                 differs:\n
                                 Badlands: {0}\n
                                 Underworld: {1}""".format(badlands_time,
                                                           uw_time))

            airIndex = badlands_model.airIndex
            sedimentIndex = badlands_model.sedimentIndex
            XML = badlands_model.XML
            resolution = badlands_model.resolution
            checkpoint_interval = badlands_model.checkpoint_interval

            self.surfaceProcesses = surfaceProcesses.Badlands(
                airIndex, sedimentIndex,
                XML, resolution,
                checkpoint_interval,
                restartFolder=restartFolder,
                restartStep=restartStep)
            if uw.rank() == 0:
                print("Badlands restarted" + '(' + datetime.now().strftime('%Y-%m-%d %H:%M:%S') + ')')
                sys.stdout.flush()

        return

    @property
    def projMaterialField(self):
        """ Material field projected on the mesh """
        self._materialFieldProjector.solve()
        self._projMaterialField.data[:] = np.rint(
            self._projMaterialField.data[:]
        )
        return self._projMaterialField

    @property
    def projPlasticStrain(self):
        """ Plastic Strain Field projected on the mesh """
        self._plasticStrainProjector.solve()
        return self._projPlasticStrain

    @property
    def projTimeField(self):
        """ Time Field projected on the mesh """
        self._timeFieldProjector.solve()
        return self._projTimeField

    @property
    def projMeltField(self):
        """ Melt Field projected on the mesh """
        self._meltFieldProjector.solve()
        return self._projMeltField

    @property
    def strainRate_2ndInvariant(self):
        """ Strain Rate Field """
        self._strainRate_2ndInvariant = fn.tensor.second_invariant(
            self.strainRate
        )
        condition = [(self._solution_exist, self._strainRate_2ndInvariant),
                     (True, fn.misc.constant(nd(self.defaultStrainRate)))]
        self._strainRate_2ndInvariant = fn.branching.conditional(condition)
        return self._strainRate_2ndInvariant

    @property
    def strainRateField(self):
        self._strainRateField.data[:] = (
            self.strainRate_2ndInvariant.evaluate(self.mesh.subMesh))
        return self._strainRateField

    @property
    def projViscosityField(self):
        """ Viscosity Field projected on the mesh """
        self.viscosityField.data[...] = self._viscosityFn.evaluate(self.swarm)
        self._viscosityFieldProjector.solve()
        return self._projViscosityField

    @property
    def viscosityField(self):
        """ Viscosity Field on particles """
        self._viscosityField.data[:] = self._viscosityFn.evaluate(self.swarm)
        return self._viscosityField

    @property
    def projStressTensor(self):
        """ Stress Tensor on mesh """
        self._stressTensor.data[...] = self._stressFn.evaluate(self.swarm)
        self._stressTensorProjector.solve()
        return self._projStressTensor

    @property
    def projStressField(self):
        """ Second Invariant of the Stress tensor projected on the submesh"""
        stress = fn.tensor.second_invariant(self._stressFn)
        self._stressField.data[...] = stress.evaluate(self.swarm)
        self._stressFieldProjector.solve()
        return self._projStressField

    @property
    def projDensityField(self):
        """ Density Field projected on the mesh """
        self.densityField.data[...] = self._densityFn.evaluate(self.swarm)
        self._densityFieldProjector.solve()
        return self._projDensityField

    @property
    def densityField(self):
        """ Density Field on particles """
        self._densityField.data[:] = self._densityFn.evaluate(self.swarm)
        return self._densityField

    @property
    def surfaceProcesses(self):
        """ Surface processes handler """
        return self._surfaceProcesses

    @surfaceProcesses.setter
    def surfaceProcesses(self, value):
        self._surfaceProcesses = value
        if value:
            self._surfaceProcesses.timeField = self.timeField
            self._surfaceProcesses.Model = self

    def set_temperatureBCs(self, left=None, right=None,
                           top=None, bottom=None,
                           front=None, back=None,
                           nodeSets=None, materials=None,
                           bottom_material=None, top_material=None,
                           left_material=None, right_material=None,
                           back_material=None, front_material=None):

        """ Set Model thermal conditions

        Parameters:
            left:
                Temperature or flux along the left wall.
                Flux must be a vector (Fx, Fy, [Fz])
                Default is 'None'
            right:
                Temperature or flux along the right wall.
                Flux must be a vector (Fx, Fy, [Fz])
                Default is 'None'
            top:
                Temperature or flux along the top wall.
                Flux must be a vector (Fx, Fy, [Fz])
                Default is 'None'
            bottom:
                Temperature or flux along the bottom wall.
                Flux must be a vector (Fx, Fy, [Fz])
                Default is 'None'
            front:
                Temperature or flux along the front wall.
                Flux must be a vector (Fx, Fy, [Fz])
                Default is 'None'
            back:
                Temperature or flux along the back wall.
                Flux must be a vector (Fx, Fy, [Fz])
                Default is 'None'
            nodeSets: (set, temperature)
                A list, numpy array which contains the indices (global)
                of the nodes and the associate temperature.
            materials:
                list of materials for which temperature need to be
                fixed. [(material, temperature)]
            bottom_material:
                Specify Material that lays at the bottom of the model
                Default is 'None' (required if a flux is defined at the
                base).
            top_material:
                Specify Material that lays at the top of the model
                Default is 'None' (required if a flux is defined at the top).
            left_material:
                Specify Material that lays at the left of the model
                Default is 'None' (required if a flux is defined at the left).
            right_material:
                Specify Material that lays at the right of the model
                Default is 'None' (required if a flux is defined at the right).
            back_material:
                Specify Material that lays at the back of the model
                Default is 'None' (required if a flux is defined at the back).
            front_material:
                Specify Material that lays at the front of the model
                Default is 'None' (required if a flux is defined at the front).

        """

        if not self.temperature:
            self.temperature = True

        self.temperatureBCs = TemperatureBCs(self, left=left, right=right,
                                             top=top, bottom=bottom,
                                             back=back, front=front,
                                             nodeSets=nodeSets,
                                             materials=materials,
                                             bottom_material=bottom_material,
                                             top_material=top_material,
                                             left_material=left_material,
                                             right_material=right_material,
                                             back_material=back_material,
                                             front_material=front_material)
        return self.temperatureBCs.get_conditions()

    @property
    def solver(self):
        return self._solver

    @solver.setter
    def solver(self, value):
        self._static_solver = True
        self._solver = value

    @property
    def temperature(self):
        """ Temperature Field """
        return self._temperature

    @temperature.setter
    def temperature(self, value):
        if value is True:
            self._temperature = MeshVariable(mesh=self.mesh,
                                                nodeDofCount=1)
            self._temperatureDot = MeshVariable(mesh=self.mesh,
                                                    nodeDofCount=1)
            self._heatFlux = MeshVariable(mesh=self.mesh,
                                              nodeDofCount=1)
            self._temperatureDot.data[...] = 0.
            self._heatFlux.data[...] = 0.
        else:
            self._temperature = False

    @property
    def _temperatureBCs(self):
        if not self.temperatureBCs:
            raise ValueError("Set Boundary Conditions")
        return self.temperatureBCs.get_conditions()

    @property
    def _advdiffSystem(self):
        """ Advection Diffusion System """

        DiffusivityMap = {}
        for material in self.materials:
            if material.diffusivity:
                DiffusivityMap[material.index] = nd(material.diffusivity)

        self.DiffusivityFn = fn.branching.map(fn_key=self.materialField,
                                              mapping=DiffusivityMap,
                                              fn_default=nd(self.diffusivity))

        HeatProdMap = {}
        for material in self.materials:
            if all([material.density,
                    material.capacity,
                    material.radiogenicHeatProd]):

                HeatProdMap[material.index] = (
                    nd(material.radiogenicHeatProd) /
                    (self._densityFn * nd(material.capacity)))
            else:
                HeatProdMap[material.index] = 0.

        self.HeatProdFn = fn.branching.map(fn_key=self.materialField,
                                           mapping=HeatProdMap)

        # Add Viscous dissipation Heating
        if rcParams["shearHeating"]:
            stress = fn.tensor.second_invariant(self._stressFn)
            strain = self.strainRate_2ndInvariant
            self.HeatProdFn += stress * strain

        if rcParams["advection.diffusion.method"] is "SLCN":

            obj = uw.systems.SLCN_AdvectionDiffusion(
                self.temperature,
                velocityField=self.velocityField,
                fn_diffusivity=self.DiffusivityFn,
                fn_sourceTerm=self.HeatProdFn,
                conditions=self._temperatureBCs
            )

        if rcParams["advection.diffusion.method"] is "SUPG":

            obj = uw.systems.AdvectionDiffusion(
                self.temperature,
                self._temperatureDot,
                velocityField=self.velocityField,
                fn_diffusivity=self.DiffusivityFn,
                fn_sourceTerm=self.HeatProdFn,
                conditions=self._temperatureBCs
            )
        return obj

    def get_stokes_solver(self):
        """ Stokes solver """

        if not self._solver or not self._static_solver:
            gravity = tuple([nd(val) for val in self.gravity])
            self._buoyancyFn = self._densityFn * gravity
            self._buoyancyFn = self._buoyancyFn

            if any([material.viscosity for material in self.materials]):

                conditions = list()
                conditions.append(self.velocityBCs.get_conditions())

                if self.stressBCs:
                    conditions.append(self.stressBCs.get_conditions())

                self._stokes_SLE = uw.systems.Stokes(
                    velocityField=self.velocityField,
                    pressureField=self.pressureField,
                    conditions=conditions,
                    fn_viscosity=self._viscosityFn,
                    fn_bodyforce=self._buoyancyFn,
                    fn_stresshistory=self._elastic_stressFn,
                    fn_one_on_lambda=self._lambdaFn)
                    #useEquationResidual=rcParams["useEquationResidual"])

                self._solver = uw.systems.Solver(self._stokes_SLE)
                self._solver.set_inner_method(rcParams["solver"])

                if rcParams["penalty"]:
                    self._solver.set_penalty(rcParams["penalty"])

                if rcParams["mg.levels"]:
                    self._solver.options.mg.levels = rcParams["mg.levels"]

            if self._solver._check_linearity(False):
                if not hasattr(self, "prevVelocityField"):
                    self.add_mesh_field("prevVelocityField", nodeDofCount=self.mesh.dim)
                if not hasattr(self, "prevPressureField"):
                    self.add_submesh_field("prevPressureField", nodeDofCount=1)

        return self._solver

    def _init_melt_fraction(self):
        """ Initialize the Melt Fraction Field """

        # Initialize Melt Fraction to material property
        meltFractionMap = {}
        for material in self.materials:
            if material.meltFraction:
                meltFractionMap[material.index] = material.meltFraction

        if meltFractionMap:
            InitFn = fn.branching.map(fn_key=self.materialField,
                                      mapping=meltFractionMap, fn_default=0.0)
            self.meltField.data[:] = InitFn.evaluate(self.swarm)

    def set_velocityBCs(self, left=None, right=None, top=None, bottom=None,
                        front=None, back=None, nodeSets=None,
                        order_wall_conditions=None):
        """ Set Model kinematic conditions

        Parameters:
            left:
                Velocity along the left wall.
                Default is 'None'
            right:
                Velocity along the right wall.
                Default is 'None'
            top:
                Velocity along the top wall.
                Default is 'None'
            bottom:
                Velocity along the bottom wall.
                Default is 'None'
            front:
                Velocity along the front wall.
                Default is 'None'
            back:
                Velocity along the back wall.
                Default is 'None'
            nodeSets: (set, velocity)
                underworld mesh index set and associate velocity
        """

        self.velocityBCs = VelocityBCs(self, left=left,
                                        right=right, top=top,
                                        bottom=bottom, front=front,
                                        back=back, nodeSets=nodeSets,
                                        order_wall_conditions=order_wall_conditions)
        return self.velocityBCs

    set_kinematicBCs = set_velocityBCs

    def set_stressBCs(self, left=None, right=None, top=None, bottom=None,
                      front=None, back=None, nodeSets=None,
                      order_wall_conditions=None):
        """ Set Model kinematic conditions

        Parameters:
            left:
                Velocity along the left wall.
                Default is 'None'
            right:
                Velocity along the right wall.
                Default is 'None'
            top:
                Velocity along the top wall.
                Default is 'None'
            bottom:
                Velocity along the bottom wall.
                Default is 'None'
            front:
                Velocity along the front wall.
                Default is 'None'
            back:
                Velocity along the back wall.
                Default is 'None'
            nodeSets: (set, velocity)
                underworld mesh index set and associate velocity
        """

        self.stressBCs = StressBCs(self, left=left,
                                   right=right, top=top,
                                   bottom=bottom, front=front,
                                   back=back, nodeSets=nodeSets,
                                   order_wall_conditions=order_wall_conditions)
        return self.stressBCs

    def add_material(self, shape=None, name="unknown", fill=True, reset=False):
        """ Add Material to the Model

        Parameters:
        -----------
            material:
                An UWGeodynamics material. If None the material is
                initialized to the global properties.
            shape:
                Shape of the material. See UWGeodynamics.shape
            name:
                Material name
            reset: (bool)
                Reset the material Field before adding the new
                material. Default is False.

        """

        if reset:
            self.materialField.data[:] = self.index

        mat = Material()
        mat.name = name
        mat.Model = self
        mat.diffusivity = self.diffusivity
        mat.capacity = self.capacity
        mat.radiogenicHeatProd = self.radiogenicHeatProd

        if isinstance(shape, shapes.Layer):
            if self.mesh.dim == 2:
                shape = shapes.Layer2D(top=shape.top, bottom=shape.bottom)
            if self.mesh.dim == 3:
                shape = shapes.Layer3D(top=shape.top, bottom=shape.bottom)

        if hasattr(shape, "top"):
            mat.top = shape.top
        if hasattr(shape, "bottom"):
            mat.bottom = shape.bottom

        mat.shape = shape
        self.materials.reverse()
        self.materials.append(mat)
        self.materials.reverse()

        if mat.shape:
            if isinstance(mat.shape, shapes.Shape):
                condition = [(mat.shape.fn, mat.index), (True, self.materialField)]
            elif isinstance(mat.shape, uw.function.Function):
                condition = [(mat.shape, mat.index), (True, self.materialField)]
            func = fn.branching.conditional(condition)
            self.materialField.data[:] = func.evaluate(self.swarm)

        return mat

    def add_swarm_field(self, name, dataType="double", count=1,
                        init_value=0., projected="mesh", **kwargs):
        """Add a new swarm field to the model

        Parameters
        ----------

        name : name of the swarm field
        dataType : type of data to be recorded, default is "double"
        count : degree of freedom, default is 1
        init_value : default value of the field, default is to initialise
            the field to 0.
        projected : the function creates a projector for each new
            swarm variable, you can choose to project on the "mesh" or
            "submesh"

        Returns
        -------

        Swarm Variable
        """

        newField = self.swarm.add_variable(dataType, count, **kwargs)
        setattr(self, name, newField)
        newField.data[...] = init_value
        self.swarm_fields[name] = newField

        # Create mesh variable for projection
        if name.startswith("_"):
            proj_name = "_proj" + name[1].upper() + name[2:]
        else:
            proj_name = "_proj" + name[0].upper() + name[1:]

        if projected == "mesh":
            projected = self.add_mesh_field(proj_name,
                                            nodeDofCount=count,
                                            dataType="double")
        else:
            projected = self.add_submesh_field(proj_name,
                                               nodeDofCount=count,
                                               dataType="double")

        # Create a projector
        if name.startswith("_"):
            projector_name = name + "Projector"
        else:
            projector_name = "_" + name + "Projector"
        projector = uw.utils.MeshVariable_Projection(projected,
                                                     newField,
                                                     voronoi_swarm=self.swarm,
                                                     type=0)
        setattr(self, projector_name, projector)

        return newField

    def add_mesh_field(self, name, nodeDofCount=1,
                       dataType="double", init_value=0., **kwargs):
        """Add a new mesh field to the model

        Parameters
        ----------

        name : name of the mesh field
        nodeDofCount : degree of freedom, default is 1
        dataType : type of data to be recorded, default is "double"
        init_value : default value of the field, default is to initialise
            the field to 0.

        Returns
        -------
        Mesh Variable
        """
        newField = self.mesh.add_variable(nodeDofCount, dataType, **kwargs)
        setattr(self, name, newField)
        newField.data[...] = init_value
        self.mesh_fields[name] = newField
        return newField

    def add_submesh_field(self, name, nodeDofCount=1,
                          dataType="double", init_value=0., **kwargs):
        """Add a new sub-mesh field to the model

        Parameters
        ----------

        name : name of the submesh field
        nodeDofCount : degree of freedom, default is 1
        dataType : type of data to be recorded, default is "double"
        init_value : default value of the field, default is to initialise
            the field to 0.

        Returns
        -------
        Mesh Variable
        """
        newField = MeshVariable(self.mesh.subMesh, nodeDofCount,
                                dataType, **kwargs)
        setattr(self, name, newField)
        newField.data[...] = init_value
        self.submesh_fields[name] = newField
        return newField

    @property
    def _densityFn(self):
        """Density Function Builder"""
        densityMap = {}
        for material in self.materials:

            if self.temperature:
                dens_handler = material.density
                dens_handler.temperatureField = self.temperature
                dens_handler.pressureField = self.pressureField
                densityMap[material.index] = dens_handler.effective_density()
            else:
                dens_handler = material.density
                densityMap[material.index] = nd(dens_handler.reference_density)

            if material.meltExpansion:
                fact = material.meltExpansion * self.meltField
                densityMap[material.index] = (
                    densityMap[material.index] * (1.0 - fact))

        return fn.branching.map(fn_key=self.materialField, mapping=densityMap)

    def set_frictional_boundary(self, right=None, left=None,
                                top=None, bottom=None, front=None,
                                back=None, thickness=2):
        """ Set Frictional Boundary conditions

        Frictional boundaries are implemented as a thin layer of frictional
        material along the chosen boundaries.

        Parameters:
        -----------
        Returns:
        --------
            Underworld mesh variable that maps the boundaries.
            (Values are set to 1 if the mesh element applies
            a frictional condition, 0 otherwise).
        """

        self.frictionalBCs = FrictionBoundaries(self, rightFriction=right,
                                                leftFriction=left,
                                                topFriction=top,
                                                bottomFriction=bottom,
                                                frontFriction=front,
                                                backFriction=back,
                                                thickness=thickness)

        return self.frictionalBCs

    @property
    def _viscosityFn(self):
        """ Viscosity Function Builder"""

        ViscosityMap = {}
        BGViscosityMap = {}

        # Viscous behavior
        for material in self.materials:
            if material.viscosity:
                ViscosityHandler = material.viscosity
                ViscosityHandler.pressureField = self.pressureField
                ViscosityHandler.strainRateInvariantField = (
                    self.strainRate_2ndInvariant)
                ViscosityHandler.temperatureField = self.temperature
                ViscosityMap[material.index] = ViscosityHandler.muEff

        # Elasticity
        for material in self.materials:
            if material.elasticity:

                # We are dealing with viscous material
                # Raise an error is no viscosity has been defined.
                if not material.viscosity:
                    raise ValueError("""Viscosity undefined for
                                     {0}""".format(material.name))
                ElasticityHandler = material.elasticity
                ElasticityHandler.viscosity = ViscosityMap[material.index]
                ViscosityMap[material.index] = ElasticityHandler.muEff

        # Melt Modifier
        for material in self.materials:
            if material.melt:
                X1 = material.viscosityChangeX1
                X2 = material.viscosityChangeX2
                change = (1.0 + (material.viscosityChange - 1.0)
                          / (X2 - X1) * (self.meltField - X1))
                conditions = [(self.meltField < X1, 1.0),
                              (self.meltField > X2, material.viscosityChange),
                              (True, change)]
                ViscosityMap[material.index] *= fn.branching.conditional(
                    conditions)


        # Plasticity
        PlasticityMap = {}
        for material in self.materials:
            if material.plasticity:

                YieldHandler = material.plasticity
                YieldHandler.pressureField = self.pressureField
                YieldHandler.plasticStrain = self.plasticStrain

                if self.mesh.dim == 2:
                    yieldStress = YieldHandler._get_yieldStress2D()

                if self.mesh.dim == 3:
                    yieldStress = YieldHandler._get_yieldStress3D()

                if material.stressLimiter:
                    stressLimiter = StressLimiter(material.stressLimiter)
                    yieldStress = stressLimiter.apply(yieldStress)
                elif self.stressLimiter:
                    stressLimiter = StressLimiter(self.stressLimiter)
                    yieldStress = stressLimiter.apply(yieldStress)

                if material.elasticity:
                    ElasticityHandler = material.elasticity
                    mu = nd(ElasticityHandler.shear_modulus)
                    dt_e = nd(ElasticityHandler.observation_time)
                    strainRate = fn.tensor.symmetric(
                        self.velocityField.fn_gradient)
                    D_eff = (strainRate
                             + 0.5
                             * self._previousStressField
                             / (mu * dt_e))
                    SRInv = fn.tensor.second_invariant(D_eff)
                else:
                    SRInv = self.strainRate_2ndInvariant

                eij = fn.branching.conditional(
                    [(SRInv < nd(1e-20 / u.second), nd(1e-20 / u.second)),
                     (True, SRInv)])

                muEff = 0.5 * yieldStress / eij
                PlasticityMap[material.index] = muEff

            if self.frictionalBCs is not None:
                from copy import copy

                # Only affect plastic materials
                if material.plasticity:

                    YieldHandler = copy(material.plasticity)
                    YieldHandler.frictionCoefficient = self.frictionalBCs.friction
                    YieldHandler.frictionAfterSoftening = self.frictionalBCs.friction
                    YieldHandler.pressureField = self.pressureField
                    YieldHandler.plasticStrain = self.plasticStrain

                    if self.mesh.dim == 2:
                        yieldStress = YieldHandler._get_yieldStress2D()

                    if self.mesh.dim == 3:
                        yieldStress = YieldHandler._get_yieldStress3D()

                    eij = fn.branching.conditional(
                        [(self.strainRate_2ndInvariant <= 1e-20, 1e-20),
                         (True, self.strainRate_2ndInvariant)])

                    muEff = 0.5 * yieldStress / eij

                    conditions = [(self.frictionalBCs._mask > 0.0, muEff),
                                  (True, PlasticityMap[material.index])]

                    PlasticityMap[material.index] = (
                        fn.branching.conditional(conditions)
                    )

        # Combine rheologies
        EffViscosityMap = {}
        PlasticMap = {}
        for material in self.materials:
            idx = material.index
            if material.viscosity and material.plasticity:
                EffViscosityMap[idx] = fn.misc.min(PlasticityMap[idx],
                                                   ViscosityMap[idx])
                BGViscosityMap[idx] = ViscosityMap[idx]
                PlasticMap[idx] = 0.
            elif material.viscosity:
                EffViscosityMap[idx] = ViscosityMap[idx]
                BGViscosityMap[idx] = ViscosityMap[idx]
                PlasticMap[idx] = 0.
            elif material.plasticity:
                EffViscosityMap[idx] = PlasticityMap[idx]
                BGViscosityMap[idx] = PlasticityMap[idx]
                PlasticMap[idx] = 1.0

        # Apply viscosity Limiter
        for material in self.materials:
            idx = material.index
            if not material.minViscosity:
                minViscosity = self.minViscosity
            else:
                minViscosity = material.minViscosity

            if not material.maxViscosity:
                maxViscosity = self.maxViscosity
            else:
                maxViscosity = material.maxViscosity

            viscosity_limiter = ViscosityLimiter(minViscosity,
                                                 maxViscosity)

            if material.viscosity or material.plasticity:
                EffViscosityMap[idx] = (
                    viscosity_limiter.apply(EffViscosityMap[idx])
                )

        viscosityFn = fn.branching.map(fn_key=self.materialField,
                                       mapping=EffViscosityMap)

        backgroundViscosityFn = fn.branching.map(fn_key=self.materialField,
                                                 mapping=BGViscosityMap)

        isPlastic = fn.branching.map(fn_key=self.materialField,
                                     mapping=PlasticMap)
        yieldConditions = [(viscosityFn < backgroundViscosityFn, 1.0),
                           (isPlastic > 0.5, 1.0),
                           (True, 0.0)]

        # Do not yield at the very first solve
        if self._solution_exist.value:
            self._isYielding = (fn.branching.conditional(yieldConditions)
                                * self.strainRate_2ndInvariant)
        else:
            self._isYielding = fn.misc.constant(0.0)

        return viscosityFn

    @property
    def _stressFn(self):
        """Stress Function Builder"""
        stressMap = {}
        for material in self.materials:
            # Calculate viscous stress
            viscousStressFn = self._viscous_stressFn()
            elasticStressFn = self._elastic_stressFn
            stressMap[material.index] = viscousStressFn + elasticStressFn

        return fn.branching.map(fn_key=self.materialField,
                                mapping=stressMap)

    def _viscous_stressFn(self):
        """Viscous Stress Function Builder"""
        return 2. * self._viscosityFn * self.strainRate

    @property
    def _elastic_stressFn(self):
        """ Elastic Stress Function Builder"""
        if any([material.elasticity for material in self.materials]):
            stressMap = {}
            for material in self.materials:
                if material.elasticity:
                    ElasticityHandler = material.elasticity
                    ElasticityHandler.viscosity = self._viscosityFn
                    ElasticityHandler.previousStress = self._previousStressField
                    elasticStressFn = ElasticityHandler.elastic_stress
                else:
                    # Set elastic stress to zero if material
                    # has no elasticity
                    elasticStressFn = [0.0] * 3 if self.mesh.dim == 2 else [0.0] * 6
                stressMap[material.index] = elasticStressFn
            return (fn.branching.map(fn_key=self.materialField,
                                     mapping=stressMap))
        else:

            elasticStressFn = [0.0] * 3 if self.mesh.dim == 2 else [0.0] * 6
            return elasticStressFn

    def _update_stress_history(self, dt):
        """Update Previous Stress Field"""
        dt_e = []
        for material in self.materials:
            if material.elasticity:
                dt_e.append(nd(material.elasticity.observation_time))
        dt_e = np.array(dt_e).min()
        phi = dt / dt_e
        veStressFn_data = self._stressFn.evaluate(self.swarm)
        self._previousStressField.data[:] *= (1. - phi)
        self._previousStressField.data[:] += phi * veStressFn_data[:]

    @property
    def _yieldStressFn(self):
        """ Calculate Yield stress function from viscosity and strain rate"""
        eij = self.strainRate_2ndInvariant
        eijdef = nd(self.defaultStrainRate)
        return 2.0 * self._viscosityFn * fn.misc.max(eij, eijdef)

    def _phaseChangeFn(self):
        for material in self.materials:
            if material.phase_changes:
                for change in material.phase_changes:
                    obj = change
                    mask = obj.fn().evaluate(self.swarm)
                    conds = ((mask == 1) &
                             (self.materialField.data == material.index))
                    self.materialField.data[conds] = obj.result

    def solve_temperature_steady_state(self):
        """ Solve for steady state temperature

        Returns:
        --------
            Updated temperature Field
        """

        if self.materials:

            DiffusivityMap = {}
            for material in self.materials:
                if material.diffusivity:
                    DiffusivityMap[material.index] = nd(material.diffusivity)

            self.DiffusivityFn = fn.branching.map(
                fn_key=self.materialField,
                mapping=DiffusivityMap,
                fn_default=nd(self.diffusivity)
            )

            HeatProdMap = {}
            for material in self.materials:

                if all([material.density,
                        material.capacity,
                        material.radiogenicHeatProd]):

                    HeatProdMap[material.index] = (
                        nd(material.radiogenicHeatProd) /
                        self._densityFn  /
                        nd(material.capacity)
                    )

                else:
                    HeatProdMap[material.index] = 0.

                # Melt heating
                if material.latentHeatFusion and self._dt:
                    dynamicHeating = self._get_dynamic_heating(material)
                    HeatProdMap[material.index] += dynamicHeating

            self.HeatProdFn = fn.branching.map(fn_key=self.materialField,
                                               mapping=HeatProdMap)
        else:
            self.DiffusivityFn = fn.misc.constant(nd(self.diffusivity))
            self.HeatProdFn = fn.misc.constant(nd(self.radiogenicHeatProd))

        conditions = self._temperatureBCs

        heatequation = uw.systems.SteadyStateHeat(
            temperatureField=self.temperature,
            fn_diffusivity=self.DiffusivityFn,
            fn_heating=self.HeatProdFn,
            conditions=conditions
        )

        heatsolver = uw.systems.Solver(heatequation)
        heatsolver.solve(nonLinearIterate=True)

        return self.temperature

    def get_lithostatic_pressureField(self):
        """ Calculate the lithostatic Pressure Field

        Returns:
        --------
            Pressure Field, array containing the pressures
            at the bottom of the Model
        """

        gravity = np.abs(nd(self.gravity[-1]))
        lithoPress = LithostaticPressure(self.mesh, self._densityFn, gravity)
        self.pressureField.data[:], LPressBot = lithoPress.solve()
        self.pressSmoother.smooth()

        return self.pressureField, LPressBot

    def _calibrate_pressureField(self):
        """ Pressure Calibration callback function """

        surfaceArea = uw.utils.Integral(fn=1.0, mesh=self.mesh,
                                        integrationType='surface',
                                        surfaceIndexSet=self.top_wall)

        surfacepressureFieldIntegral = uw.utils.Integral(
            fn=self.pressureField,
            mesh=self.mesh,
            integrationType='surface',
            surfaceIndexSet=self.top_wall
        )
        area, = surfaceArea.evaluate()
        p0, = surfacepressureFieldIntegral.evaluate()
        offset = p0 / area
        self.pressureField.data[:] -= offset

    def _get_material_indices(self, material):
        """ Get mesh indices of a Material

        Parameter:
        ----------
            material:
                    The Material of interest
        Returns;
        --------
            underworld IndexSet

        """
        mat = self.projMaterialField.evaluate(self.mesh.data[0:self.mesh.nodesLocal])
        # Round to closest integer
        mat = np.rint(mat)
        # convert array as integer
        mat = mat.astype("int")[:, 0]
        # Get nodes corresponding to material
        nodes = np.arange(0, self.mesh.nodesLocal)[mat == material.index]
        return uw.mesh.FeMesh_IndexSet(self.mesh, topologicalIndex=0,
                                       size=self.mesh.nodesGlobal,
                                       fromObject=nodes)

    def solve(self):
        """ Solve Stokes """

        if self.step == 0:
            self._curTolerance = rcParams["initial.nonlinear.tolerance"]
            minIterations = rcParams["initial.nonlinear.min.iterations"]
            maxIterations = rcParams["initial.nonlinear.max.iterations"]
        else:
            self._curTolerance = rcParams["nonlinear.tolerance"]
            minIterations = rcParams["nonlinear.min.iterations"]
            maxIterations = rcParams["nonlinear.max.iterations"]

        self.get_stokes_solver().solve(
            nonLinearIterate=True,
            nonLinearMinIterations=minIterations,
            nonLinearMaxIterations=maxIterations,
            callback_post_solve=self.callback_post_solve,
            nonLinearTolerance=self._curTolerance)

        self._solution_exist.value = True

    def init_model(self, temperature=True, pressureField=True):
        """ Initialize the Temperature Field as steady state,
            Initialize the Pressure Field as Lithostatic

        Parameters:
        -----------
            temperature: (bool) default to True
            pressure: (bool) default to True
        """

        # Init Temperature Field
        if self.temperature and temperature:
            self.solve_temperature_steady_state()

        # Init pressureField Field
        if self.pressureField and pressureField:
            self.get_lithostatic_pressureField()
        return

    @u.check([None, "[time]", "[time]", None, None, None, "[time]", None, None])
    def run_for(self, duration=None, checkpoint_interval=None, nstep=None,
                checkpoint_times=None, restart_checkpoint=1, dt=None,
                restartStep=-1, restartDir=None):
        """ Run the Model

        Parameters
        ----------

        duration :
            Model time in units of time.
        checkpoint_interval :
            Checkpoint interval time.
        nstep :
            Number of steps to run.`
        checkpoint_times :
            Specify a list of additional Checkpoint times ([Time])
        restart_checkpoint :
            This parameter specify how often the swarm and swarm variables
            are checkpointed. A value of 1 means that the swarm and its
            associated variables are saved at every checkpoint.
            A value of 2 results in saving only every second checkpoint.
        dt :
            Specify the time interval (dt) to be used in
            units of time.
        restartStep :
            Restart Model. int (step number)
        restartDir :
            Restart Directory.

        """

        if uw.rank() == 0:
            print("""Running with UWGeodynamics version {0}""".format(full_version))
            sys.stdout.flush()

        if restartStep:
            restartDir = restartDir if restartDir else self.outputDir
            if os.path.exists(restartDir):
                self.restart(step=restartStep, restartDir=restartDir)
            uw.barrier()

        if ((checkpoint_interval or checkpoint_times) and
            uw.rank() == 0 and not os.path.exists(self.outputDir)):
            os.makedirs(self.outputDir)
        uw.barrier()

        stepDone = 0
        time = nd(self.time)

        if duration:
            units = duration.units
            duration = time + nd(duration)
        else:
            units = rcParams["time.SIunits"]
            duration = time

        if not nstep:
            nstep = stepDone

        if dt:
            user_dt = nd(dt)
        else:
            user_dt = None

        next_checkpoint = None

        if checkpoint_times:
            checkpoint_times = [nd(val) for val in checkpoint_times]

        if (isinstance(checkpoint_interval, u.Quantity) and
           checkpoint_interval.dimensionality == _dim_time):
            next_checkpoint = time + nd(checkpoint_interval)
        elif checkpoint_interval:
            next_checkpoint = stepDone + checkpoint_interval

        # Save initial state
        if checkpoint_interval or checkpoint_times:
            self.checkpoint()

        while time < duration or stepDone < nstep:

            self.preSolveHook()

            self.solve()

            # Whats the longest we can run before reaching the end
            # of the model or a checkpoint?
            # Need to generalize that
            self._dt = 2.0 * rcParams["CFL"] * self.swarm_advector.get_max_dt()

            if self.temperature:
                # Only get a condition if using SUPG
                if rcParams["advection.diffusion.method"] == "SUPG":
                    supg_dt = self._advdiffSystem.get_max_dt()
                    supg_dt *= 2.0 * rcParams["CFL"]
                    self._dt = min(self._dt, supg_dt)

            if (isinstance(checkpoint_interval, u.Quantity) and
               checkpoint_interval.dimensionality == _dim_time):
                self._dt = min(self._dt, next_checkpoint - time)

            if duration:
                self._dt = min(self._dt, duration - time)

            if user_dt:
                self._dt = min(self._dt, user_dt)

            if checkpoint_times:
                tcheck = [val for val in (checkpoint_times - time) if val >= 0]
                tcheck.sort()
                self._dt = min(self._dt, tcheck[0])

            dte = []
            for material in self.materials:
                if material.elasticity:
                    dte.append(nd(material.elasticity.observation_time))

            if dte:
                dte = np.array(dte).min()
                # Cap dt for observation time, dte / 3.
                if dte and self._dt > (dte / 3.):
                    self._dt = dte / 3.

            uw.barrier()

            self._update()

            self.step += 1
            stepDone += 1
            self.time += Dimensionalize(self._dt, units)
            time += self._dt

            if ((isinstance(checkpoint_interval, u.Quantity) and
               checkpoint_interval.dimensionality == _dim_time and
               time == next_checkpoint) or stepDone == next_checkpoint):
                    self.checkpointID += 1
                    # Save Mesh Variables
                    self.checkpoint_fields(checkpointID=self.checkpointID)
                    # Save Tracers
                    self.checkpoint_tracers(checkpointID=self.checkpointID)
                    next_checkpoint += nd(checkpoint_interval)

            uw.barrier()

            # if it's time to checkpoint the swarm, do so.
            if self.checkpointID % restart_checkpoint == 0:
                self.checkpoint_swarms(checkpointID=self.checkpointID)

            uw.barrier()

            if checkpoint_interval or self.step % 1 == 0 or nstep:
                if uw.rank() == 0:
                    print("Step:" + str(stepDone) + " Model Time: ", str(self.time.to(units)),
                          'dt:', str(Dimensionalize(self._dt, units)),
                          '(' + datetime.now().strftime('%Y-%m-%d %H:%M:%S') + ')')
                    sys.stdout.flush()

            self.postSolveHook()

        return 1

    @staticmethod
    def preSolveHook():
        """ Entry point for functions to be run before attempting a solve """
        pass

    @staticmethod
    def postSolveHook():
        """ Entry point for functions to be run after the solve """
        pass

    @property
    def callback_post_solve(self):
        """ Function called right after a non-linear iteration or a linear
        solve"""
        return self._callback_post_solve

    @callback_post_solve.setter
    def callback_post_solve(self, value):
        def callback():
            if callable(value):
                value()
            if rcParams["surface.pressure.normalization"]:
                self._calibrate_pressureField()
            if rcParams["pressure.smoothing"]:
                self.pressSmoother.smooth()
            if self._isostasy:
                self._isostasy.solve()
            for material in self.materials:
                if material.viscosity:
                    material.viscosity.firstIter.value = False

            self._solution_exist.value = True
        self._callback_post_solve = callback

    def _update(self):
        """ Update Function

        The following function processes the mesh and swarm variables
        between two solves. It takes care of mesh, swarm advection and
        update the fields according to the Model state.

        """

        dt = self._dt

        # Heal plastic strain
        if any([material.healingRate for material in self.materials]):
            healingRates = {}
            for material in self.materials:
                healingRates[material.index] = nd(material.healingRate)
            HealingRateFn = fn.branching.map(fn_key=self.materialField,
                                             mapping=healingRates)

            plasticStrainIncHealing = dt * HealingRateFn.evaluate(self.swarm)
            self.plasticStrain.data[:] -= plasticStrainIncHealing
            self.plasticStrain.data[self.plasticStrain.data < 0.] = 0.

        # Increment plastic strain
        plasticStrainIncrement = dt * self._isYielding.evaluate(self.swarm)
        self.plasticStrain.data[:] += plasticStrainIncrement

        if any([material.melt for material in self.materials]):
            # Calculate New meltField
            self.update_melt_fraction()

        # Solve for temperature
        if self.temperature:
            self._advdiffSystem.integrate(dt)

        if self._advector:
            self.swarm_advector.integrate(dt)
            self._advector.advect_mesh(dt)
        elif self._freeSurface:
            self.swarm_advector.integrate(dt, update_owners=False)
            self._freeSurface.solve(dt)
            self.swarm.update_particle_owners()
        else:
            # Integrate Swarms in time
            self.swarm_advector.integrate(dt, update_owners=True)

        # Update stress
        if any([material.elasticity for material in self.materials]):
            self._update_stress_history(dt)

        if self.passive_tracers:
            for key in self.passive_tracers:
                self.passive_tracers[key].integrate(dt)

        # Do pop control
        self.population_control.repopulate()
        self.swarm.update_particle_owners()

        if self.surfaceProcesses:
            self.surfaceProcesses.solve(dt)

        # Update Time Field
        self.timeField.data[...] += dt


        if self._visugrid:
            self._visugrid.advect(dt)

        self._phaseChangeFn()

    def mesh_advector(self, axis):
        """ Initialize the mesh advector

        Parameters:
        -----------
            axis:
                list of axis (or degree of freedom) along which the
                mesh is allowed to deform
        """
        self._advector = _mesh_advector(self, axis)

    def add_passive_tracers(self, name, vertices=None,
                            particleEscape=True, centroids=None):
        """ Add a swarm of passive tracers to the Model

        Parameters:
        -----------
            name :
                Name of the swarm of tracers.
            vertices :
                Numpy array that contains the coordinates of the tracers.
            particleEscape : (bool)
                Allow or prevent tracers from escaping the boundaries of the
                Model (default to True)
            centroids : if a list of centroids is provided, the pattern defined
                by the vertices is reproduced around each centroid.
        """

        if centroids and not isinstance(centroids, list):
            centroids = list(centroids)

        if name in self.passive_tracers.keys():
            print("{0} tracers exists already".format(name))
            sys.stdout.flush()
            return self.passive_tracers[name]

        if not centroids:

            tracers = PassiveTracers(self.mesh,
                                     self.velocityField,
                                     name=name,
                                     vertices=vertices,
                                     particleEscape=particleEscape)

        else:

            tracers = PassiveTracersGrid(self.mesh,
                                         self.velocityField,
                                         name=name,
                                         vertices=vertices,
                                         centroids=centroids,
                                         particleEscape=particleEscape)

        self.passive_tracers[name] = tracers
        setattr(self, name.lower() + "_tracers", tracers)

        return tracers

    def _get_melt_fraction(self):
        """ Melt Fraction function

        Returns:
        -------
            Underworld function that calculates the Melt fraction on the
            particles.

        """
        meltMap = {}
        for material in self.materials:
            if material.melt:
                T_s = material.solidus.temperature(self.pressureField)
                T_l = material.liquidus.temperature(self.pressureField)
                T_ss = (self.temperature - 0.5 * (T_s + T_l)) / (T_l - T_s)
                value = (0.5 + T_ss + (T_ss * T_ss - 0.25) *
                         (0.4256 + 2.988 * T_ss))
                conditions = [((-0.5 < T_ss) & (T_ss < 0.5),
                               fn.misc.min(value, material.meltFractionLimit)),
                              (True, 0.0)]
                meltMap[material.index] = fn.branching.conditional(conditions)

        return fn.branching.map(fn_key=self.materialField,
                                mapping=meltMap, fn_default=0.0)

    def update_melt_fraction(self):
        """ Calculate New meltField """
        meltFraction = self._get_melt_fraction()
        self.meltField.data[:] = meltFraction.evaluate(self.swarm)

    def _get_dynamic_heating(self, material):
        """ Calculate additional heating source due to melt

        Returns:
        --------
            Underworld function

        """

        ratio = material.latentHeatFusion / material.capacity

        if not ratio.dimensionless:
            raise ValueError("""Unit Error in either Latent Heat Fusion or
                             Capacity (Material: """ + material.name)
        ratio = ratio.magnitude

        dF = (self._get_melt_fraction() - self.meltField) / self._dt
        return (ratio * dF) * self.temperature

    @property
    def _lambdaFn(self):
        """ Initialize compressibility """

        materialMap = {}
        if any([material.compressibility for material in self.materials]):
            for material in self.materials:
                if material.compressibility:
                    materialMap[material.index] = nd(material.compressibility)

            return (uw.function.branching.map(fn_key=self.materialField,
                                                  mapping=materialMap,
                                                  fn_default=0.0))
        return

    @property
    def freeSurface(self):
        return self._freeSurface

    @freeSurface.setter
    def freeSurface(self, value):
        if value:
            self._freeSurface = FreeSurfaceProcessor(self)

    def checkpoint(self, checkpointID=None, variables=None):
        """ Do a checkpoint (Save fields)

        Parameters:
        -----------
            variables:
                list of fields/variables to save
            checkpointID:
                checkpoint ID.

        """

        self.checkpoint_fields(variables, checkpointID)
        self.checkpoint_swarms(variables, checkpointID)
        self.checkpoint_tracers(checkpointID=checkpointID)

    def add_visugrid(self, elementRes, minCoord=None, maxCoord=None):
        """ Add a tracking grid to the Model

        This is essentially a lagrangian grid that deforms with the materials.

        Parameters:
        -----------
            elementRes:
                Grid resolution in number of elements
                along each axis (x, y, z).
            minCoord:
                Minimum coordinate for each axis.
                Used to define the extent of the grid,
            maxCoord:
                Maximum coordinate for each axis.
                Used to define the extent of the grid,
        """

        if not maxCoord:
            maxCoord = self.maxCoord

        if not minCoord:
            minCoord = self.minCoord

        self._visugrid = Visugrid(self, elementRes, minCoord, maxCoord,
                                  self.velocityField)

    def checkpoint_fields(self, fields=None, checkpointID=None, time=None,
                          outputDir=None):
        """ Save the mesh and the mesh variables to outputDir

        Parameters
        ----------

        fields : A list of mesh/field variables to be saved.
        checkpointID : Checkpoint ID
        time : Model time at checkpoint
        outputDir : output directory

        """

        if not fields:
            fields = rcParams["default.outputs"]

        if not checkpointID:
            checkpointID = self.checkpointID

        if not outputDir:
            outputDir = self.outputDir

        if uw.rank() == 0 and not os.path.exists(outputDir):
            os.makedirs(outputDir)
        uw.barrier()

        time = time if time else self.time

        if self._advector or self._freeSurface:
            mesh_name = 'mesh-%s' % checkpointID
            mesh_prefix = os.path.join(outputDir, mesh_name)
            mH = self.mesh.save('%s.h5' % mesh_prefix, units=u.kilometers,
                                time=time)
        elif not self._mesh_saved:
            mesh_name = 'mesh'
            mesh_prefix = os.path.join(outputDir, mesh_name)
            mH = self.mesh.save('%s.h5' % mesh_prefix, units=u.kilometers,
                                time=time)
            self._mesh_saved = True
        else:
            mesh_name = 'mesh'
            mesh_prefix = os.path.join(outputDir, mesh_name)
            mH = uw.utils.SavedFileData(self.mesh, '%s.h5' % mesh_prefix)

        if uw.rank() == 0:
            filename = "XDMF.fields." + str(checkpointID).zfill(5) + ".xmf"
            filename = os.path.join(outputDir, filename)

            # First write the XDMF header
            string = uw.utils._xdmfheader()
            string += uw.utils._spacetimeschema(mH, mesh_name, time)

        uw.barrier()

        for field in fields:
            if field == "temperature" and not self.temperature:
                continue
            if field in rcParams["mesh.variables"]:
                field = str(field)
                try:
                    units = rcParams[field + ".SIunits"]
                except KeyError:
                    units = None

                # Save the h5 file and write the field schema for
                # each one of the field variables
                obj = getattr(self, field)
                file_prefix = os.path.join(outputDir, field + '-%s' % checkpointID)
                handle = obj.save('%s.h5' % file_prefix, units=units,
                                  time=time)
                if uw.rank() == 0:
                    string += uw.utils._fieldschema(handle, field)
            uw.barrier()

        if uw.rank() == 0:
            # Write the footer to the xmf
            string += uw.utils._xdmffooter()

            # Write the string to file - only proc 0
            with open(filename, "w") as xdmfFH:
                xdmfFH.write(string)
        uw.barrier()

    def checkpoint_swarms(self, fields=None, checkpointID=None, time=None,
                          outputDir=None):
        """ Save the swarm and the swarm variables to outputDir

        Parameters
        ----------

        fields : A list of swarm/field variables to be saved.
        checkpointID : Checkpoint ID
        time : Model time at checkpoint
        outputDir : output directory

        """

        if not fields:
            fields = rcParams["restart.fields"]

        if not checkpointID:
            checkpointID = self.checkpointID

        if not outputDir:
            outputDir = self.outputDir

        if uw.rank() == 0 and not os.path.exists(outputDir):
            os.makedirs(outputDir)
        uw.barrier()

        time = time if time else self.time
        swarm_name = 'swarm-%s.h5' % checkpointID

        sH = self.swarm.save(os.path.join(outputDir,
                             swarm_name),
                             units=u.kilometers,
                             time=time)

        if uw.rank() == 0:
            filename = "XDMF.swarms." + str(checkpointID).zfill(5) + ".xmf"
            filename = os.path.join(outputDir, filename)

            # First write the XDMF header
            string = uw.utils._xdmfheader()
            string += uw.utils._swarmspacetimeschema(sH, swarm_name, time)
        uw.barrier()

        for field in fields:
            if field in rcParams["swarm.variables"]:
                field = str(field)
                try:
                    units = rcParams[field + ".SIunits"]
                except KeyError:
                    units = None

                # Save the h5 file and write the field schema for
                # each one of the field variables
                obj = getattr(self, field)
                file_prefix = os.path.join(outputDir, field + '-%s' % checkpointID)
                handle = obj.save('%s.h5' % file_prefix, units=units, time=time)
                if uw.rank() == 0:
                    string += uw.utils._swarmvarschema(handle, field)
                uw.barrier()

        if uw.rank() == 0:
            # Write the footer to the xmf
            string += uw.utils._xdmffooter()

            # Write the string to file - only proc 0
            with open(filename, "w") as xdmfFH:
                xdmfFH.write(string)

        uw.barrier()

    @u.check([None, None, None, "[time]", None])
    def checkpoint_tracers(self, tracers=None, checkpointID=None,
                           time=None, outputDir=None):
        """ Checkpoint the tracers

        Parameters
        ----------

        tracers : List of tracers to checkpoint.
        checkpointID : Checkpoint ID.
        time : Model time at checkpoint.
        outputDir : output directory

        """

        if not checkpointID:
            checkpointID = self.checkpointID

        time = time if time else self.time

        if not outputDir:
            outputDir = self.outputDir

        if uw.rank() == 0 and not os.path.exists(outputDir):
            os.makedirs(outputDir)
        uw.barrier()

        # Checkpoint passive tracers and associated tracked fields
        if self.passive_tracers:
            for (dump, item) in self.passive_tracers.items():
                item.save(outputDir, checkpointID, time)

    def geometry_from_shapefile(self, filename, units=None):
        from ._utils import MoveImporter
        Importer = MoveImporter(filename, units=units)

        shape_dict = {name: [] for name in Importer.names}

        for polygon in Importer.generator:
            name = polygon["properties"]["Name"]
            vertices = polygon["coordinates"]
            shape = shapes.Polygon(vertices)
            shape_dict[name].append(shape)

        for name, shape in shape_dict.items():
            mshape = shapes.MultiShape(shape)
            self.add_material(name=name,
                              shape=mshape,
                              fill=False)

        self._fill_model()

    def velocity_rms(self):
        vdotv_fn = uw.function.math.dot(self.velocityField, self.velocityField)
        fn_2_integrate = (1., vdotv_fn)
        (v2, vol) = self.mesh.integrate(fn=fn_2_integrate)
        import math
        vrms = math.sqrt(v2 / vol)
        #os.write(1, "Velocity rms (vrms): {0}".format(vrms))
        return vrms

_html_global = OrderedDict()
_html_global["Number of Elements"] = "elementRes"
_html_global["length"] = "length"
_html_global["width"] = "width"
_html_global["height"] = "height"


def _model_html_repr(Model):
    header = "<table>"
    footer = "</table>"
    html = ""
    for key, val in _html_global.items():
        value = Model.__dict__.get(val)
        html += "<tr><td>{0}</td><td>{1}</td></tr>".format(key, value)

    return header + html + footer



from __future__ import print_function,  absolute_import
import underworld as uw
import h5py
import numpy as np
from mpi4py import MPI
from UWGeodynamics.scaling import Dimensionalize
from UWGeodynamics.scaling import nonDimensionalize
from UWGeodynamics.scaling import UnitRegistry as u
from UWGeodynamics.version import git_revision as __git_revision__


class SwarmVariable(uw.swarm.SwarmVariable):

    def __init__(self, swarm, dataType, count, writeable=True, **kwargs):
        super(SwarmVariable, self).__init__(swarm, dataType, count,
                                            writeable=True, **kwargs)

    def load(self, filename, collective=False):
        """
        Load the swarm variable from disk. This must be called *after* the swarm.load().

        Parameters
        ----------
        filename : str
            The filename for the saved file. Relative or absolute paths may be
            used, but all directories must exist.

        Notes
        -----
        This method must be called collectively by all processes.


        Example
        -------
        Refer to example provided for 'save' method.

        """

        if not isinstance(filename, str):
            raise TypeError("'filename' parameter must be of type 'str'")

        if self.swarm._checkpointMapsToState != self.swarm.stateId:
            raise RuntimeError("'Swarm' associate with this 'SwarmVariable' does not appear to be in the correct state.\n" \
                               "Please ensure that you have loaded the swarm prior to loading any swarm variables.")
        gIds = self.swarm._local2globalMap

        comm = MPI.COMM_WORLD
        rank = comm.rank

        # open hdf5 file
        h5f = h5py.File(name=filename, mode="r", driver='mpio', comm=MPI.COMM_WORLD)


        dset = h5f.get('data')
        if dset == None:
            raise RuntimeError("Can't find 'data' in file '{}'.\n".format(filename))

        if dset.shape[1] != self.data.shape[1]:
            raise RuntimeError("Cannot load file data on current swarm. Data in file '{0}', " \
                               "has {1} components -the particlesCoords has {2} components".format(filename, dset.shape[1], self.particleCoordinates.data.shape[1]))

        if dset.shape[0] != self.swarm.particleGlobalCount:
            raise RuntimeError("It appears that the swarm has {} particles, but provided h5 file has {} data points. Please check that " \
                               "both the Swarm and the SwarmVariable were saved at the same time, and that you have reloaded using " \
                               "the correct files.".format(particleGobalCount, dset.shape[0]))

        # for efficiency, we want to load swarmvariable data in the largest stride chunks possible.
        # we need to determine where required data is contiguous.
        # first construct an array of gradients. the required data is contiguous
        # where the indices into the array are increasing by 1, ie have a gradient of 1.
        gradIds = np.zeros_like(gIds)            # creates array of zeros of same size & type
        if len(gIds) > 1:
            gradIds[:-1] = gIds[1:] - gIds[:-1]  # forward difference type gradient

        # note that we do only the first read into dset collective. this call usually
        # does the entire read, but if it doesn't we won't know how many calls will
        # be necessary, hence only collective calling the first.
        done_collective = False
        guy = 0
        while guy < len(gIds):
            # do contiguous
            start_guy = guy
            while gradIds[guy]==1:  # count run of contiguous. note bounds check not required as last element of gradIds is always zero.
                guy += 1
            # copy contiguous chunk if found.. note that we are copying 'plus 1' items
            if guy > start_guy:
                if collective and not done_collective:
                    with dset.collective:
                        self.data[start_guy:guy+1] = dset[gIds[start_guy]:gIds[guy]+1]
                        done_collective = True
                else:
                    self.data[start_guy:guy+1] = dset[gIds[start_guy]:gIds[guy]+1]
                guy += 1

            # do non-contiguous
            start_guy = guy
            while guy<len(gIds) and gradIds[guy]!=1:  # count run of non-contiguous
                guy += 1
            # copy non-contiguous items (if found) using index array slice
            if guy > start_guy:
                if collective and not done_collective:
                    with dset.collective:
                        self.data[start_guy:guy,:] = dset[gIds[start_guy:guy],:]
                        done_collective = True
                else:
                    self.data[start_guy:guy,:] = dset[gIds[start_guy:guy],:]

        # if we haven't entered a collective call, do so now to
        # avoid deadlock. we just do an empty read/write.
        if collective and not done_collective:
            with dset.collective:
                self.data[0:0, :] = dset[0:0, :]

        # get units
        try:
            units = h5f.attrs["units"]
        except KeyError:
            units = None

        if units and units != "None":
            units = u.parse_expression(units)
        else:
            units = None

        h5f.close()

        if units:
            self.data[:] = nonDimensionalize(self.data * units)

    def save( self, filename, collective=False, units=None, time=None):
        """
        Save the swarm variable to disk.

        Parameters
        ----------
        filename : str
            The filename for the saved file. Relative or absolute paths may be
            used, but all directories must exist.
        swarmHandle :uw.utils.SavedFileData , optional
            The saved swarm file handle. If provided, a reference to the swarm file
            is made. Currently this doesn't provide any extra functionality.

        Returns
        -------
        underworld.utils.SavedFileData
            Data object relating to saved file. This only needs to be retained
            if you wish to create XDMF files and can be ignored otherwise.

        Notes
        -----
        This method must be called collectively by all processes.

        Example
        -------
        First create the swarm, populate, then add a variable:

        >>> mesh = uw.mesh.FeMesh_Cartesian( elementType='Q1/dQ0', elementRes=(16,16), minCoord=(0.,0.), maxCoord=(1.,1.) )
        >>> swarm = uw.swarm.Swarm(mesh)
        >>> swarm.populate_using_layout(uw.swarm.layouts.PerCellGaussLayout(swarm,2))
        >>> svar = swarm.add_variable("int",1)

        Write something to variable

        >>> import numpy as np
        >>> svar.data[:,0] = np.arange(swarm.particleLocalCount)

        Save to a file:

        >>> ignoreMe = swarm.save("saved_swarm.h5")
        >>> ignoreMe = svar.save("saved_swarm_variable.h5")

        Now let's try and reload. First create a new swarm and swarm variable,
        and then load both:

        >>> clone_swarm = uw.swarm.Swarm(mesh)
        >>> clone_svar = clone_swarm.add_variable("int",1)
        >>> clone_swarm.load("saved_swarm.h5")
        >>> clone_svar.load("saved_swarm_variable.h5")

        Now check for equality:

        >>> import numpy as np
        >>> np.allclose(svar.data,clone_svar.data)
        True

        >>> # clean up:
        >>> if uw.rank() == 0:
        ...     import os;
        ...     os.remove( "saved_swarm.h5" )
        ...     os.remove( "saved_swarm_variable.h5" )

        """

        if not isinstance(filename, str):
            raise TypeError("'filename' parameter must be of type 'str'")

        # setup mpi basic vars
        comm = MPI.COMM_WORLD
        rank = comm.rank

        # allgather the number of particles each proc has
        swarm = self.swarm
        procCount = comm.allgather(swarm.particleLocalCount)
        particleGlobalCount = np.sum(procCount)

        # calculate the hdf5 file offset
        offset=0
        for i in range(comm.rank):
            offset += procCount[i]

        # open parallel hdf5 file
        with h5py.File(name=filename, mode="w", driver='mpio', comm=MPI.COMM_WORLD) as h5f:
            # write the entire local swarm to the appropriate offset position
            globalShape = (particleGlobalCount, self.data.shape[1])
            dset = h5f.create_dataset("data",
                                       shape=globalShape,
                                       dtype=self.data.dtype)
            fact = 1.0
            if units:
                fact = Dimensionalize(1.0, units=units).magnitude
                h5f.attrs['units'] = str(units)

            if time is not None:
                h5f.attrs['time'] = str(time)

            h5f.attrs["git commit"] = __git_revision__

            if collective:
                with dset.collective:
                    dset[offset:offset+swarm.particleLocalCount] = self.data[:] * fact
            else:
                dset[offset:offset + swarm.particleLocalCount] = self.data[:] * fact

        # let's reopen in serial to write the attrib.
        # not sure if this really is necessary.
        comm.barrier()
        if comm.rank==0:
            with h5py.File(name=filename, mode="a") as h5f:
                # attribute of the proc offsets - used for loading from checkpoint
                h5f.attrs["proc_offset"] = procCount

        return uw.utils.SavedFileData( self, filename )

    def copy(self, deepcopy=False):
        """
        This method returns a copy of the swarmvariable.

        Parameters
        ----------
        deepcopy: bool
            If True, the variable's data is also copied into
            new variable.

        Returns
        -------
        underworld.swarm.SwarmVariable
            The swarm variable copy.

        Example
        -------
        >>> import math
        >>> mesh = uw.mesh.FeMesh_Cartesian()
        >>> swarm = uw.swarm.Swarm()
        >>> swarm.populate.using_layout(uw.swarm.layouts.PerCellGaussLayout(swarm, 2)
        >>> svar = swarm.add_variable("double", 1)
        >>> svar.data[:] = math.exp(1.)
        >>> svarCopy = svar.copy()
        >>> svarCopy.swarm == var.swarm
        True
        >>> svarCopy.dataType == svar.dataType
        True
        >>> import numpy as np
        >>> np.allclose(svar.data,svarCopy.data)
        False
        >>> svarCopy2 = svar.copy(deepcopy=True)
        >>> np.allclose(svar.data,svarCopy2.data)
        True

        """

        if not isinstance(deepcopy, bool):
            raise TypeError("'deepcopy' parameter is expected to be of type 'bool'.")

        newSv = SwarmVariable(self.swarm, self.dataType, self.count)

        if deepcopy:
            newSv.data[:] = self.data[:]

        return newSv



from __future__ import print_function,  absolute_import
import abc
try:
    from .linkage import SPM
except ImportError:
    pass
import underworld.function as fn
from .scaling import nonDimensionalize as nd


ABC = abc.ABCMeta('ABC', (object,), {})


class SurfaceProcesses(ABC):

    def __init__(self, Model=None):

        self.Model = Model

    @property
    def Model(self):
        return self._Model

    @Model.setter
    def Model(self, value):
        self._Model = value
        if value:
            self._init_model()

    @abc.abstractmethod
    def _init_model(self):
        pass

    @abc.abstractmethod
    def solve(self, dt):
        pass


class Badlands(SurfaceProcesses):
    """ A wrapper class for Badlands Linkage"""

    def __init__(self, airIndex,
                 sedimentIndex, XML, resolution, checkpoint_interval,
                 surfElevation=0., verbose=True, Model=None, restartFolder=None,
                 restartStep=None, timeField=None, surfaceTracers=None):
        try:

            import pyBadlands

        except ImportError :
            raise ImportError("""pyBadlands import as failed. Please check your
                              installation, PYTHONPATH and PATH environment
                              variables""")

        self.airIndex = airIndex
        self.sedimentIndex = sedimentIndex
        self.XML = XML
        self.resolution = resolution
        self.checkpoint_interval = checkpoint_interval
        self.surfElevation = surfElevation
        self.verbose = verbose
        self.restartFolder = restartFolder
        self.restartStep = restartStep
        self.surfaceTracers = surfaceTracers

        self.timeField = timeField
        self.Model = Model

    def _init_model(self):
        self.mesh = self._Model.mesh
        self.velocityField = self._Model.velocityField
        self.swarm = self._Model.swarm
        self.materialField = self._Model.materialField

        self._BadlandsModel = SPM(self.mesh, self.velocityField, self.swarm,
                                  self.materialField, self.airIndex,
                                  self.sedimentIndex,
                                  self.XML, nd(self.resolution),
                                  nd(self.checkpoint_interval),
                                  nd(self.surfElevation), self.verbose,
                                  self.restartFolder, self.restartStep)
        return

    def solve(self, dt):
        self._BadlandsModel.solve(dt)
        return


class ErosionThreshold(SurfaceProcesses):

    def __init__(self, air=None, threshold=None, surfaceTracers=None,
                 Model=None, **kwargs):

        super(ErosionThreshold, self).__init__(Model=Model)

        self.Model = Model
        self.threshold = threshold
        self.air = air
        self.surfaceTracers = surfaceTracers
        self.Model = Model

    def _init_model(self):

        materialField = self.Model.materialField

        materialMap = {}
        for material in self.air:
            materialMap[material.index] = 1.0

        isAirMaterial = fn.branching.map(fn_key=materialField,
                                         mapping=materialMap,
                                         fn_default=0.0)

        belowthreshold = [(((isAirMaterial < 0.5) & (fn.input()[1] > nd(self.threshold))), self.air[0].index),
                          (True, materialField)]

        self._fn = fn.branching.conditional(belowthreshold)

    def solve(self, dt):

        if not self.Model:
            raise ValueError("Model is not defined")

        self.Model.materialField.data[:] = self._fn.evaluate(self.Model.swarm)
        if self.surfaceTracers:
            if self.surfaceTracers.swarm.particleCoordinates.data.size > 0:
                coords = self.surfaceTracers.swarm.particleCoordinates
                coords.data[coords.data[:, -1] > nd(self.threshold), -1] = nd(self.threshold)
        return


class SedimentationThreshold(SurfaceProcesses):

    def __init__(self, air=None, sediment=None,
                 threshold=None, timeField=None, Model=None,
                 surfaceTracers=None, **kwargs):

        super(SedimentationThreshold, self).__init__(Model=Model)

        self.timeField = timeField
        self.air = air
        self.sediment = sediment
        self.threshold = threshold
        self.surfaceTracers = surfaceTracers
        self.Model = Model

    def _init_model(self):

        materialField = self.Model.materialField

        materialMap = {}
        for material in self.air:
            materialMap[material.index] = 1.0

        isAirMaterial = fn.branching.map(fn_key=materialField,
                                         mapping=materialMap,
                                         fn_default=0.0)

        belowthreshold = [(((isAirMaterial > 0.5) & (fn.input()[1] < nd(self.threshold))), 0.),
                          (True, 1.)]

        self._change_material = fn.branching.conditional(belowthreshold)

        conditions = [(self._change_material < 0.5, self.sediment[0].index),
                      (True, materialField)]

        self._fn = fn.branching.conditional(conditions)

    def solve(self, dt):

        if not self.Model:
            raise ValueError("Model is not defined")

        if self.timeField:
            fn = self._change_material * self.timeField
            self.timeField.data[...] = fn.evaluate(self.Model.swarm)

        self.Model.materialField.data[:] = self._fn.evaluate(self.Model.swarm)

        if self.surfaceTracers:
            if self.surfaceTracers.swarm.particleCoordinates.data.size > 0:
                coords = self.surfaceTracers.swarm.particleCoordinates
                coords.data[coords.data[:, -1] < nd(self.threshold), -1] = nd(self.threshold)


class ErosionAndSedimentationThreshold(SedimentationThreshold, ErosionThreshold):

    def __init__(self, air=None, sediment=None,
                 threshold=None, timeField=None,
                 surfaceTracers=None, Model=None, **kwargs):

        super(ErosionAndSedimentationThreshold, self).__init__(Model=Model)

        self.timeField = timeField
        self.air = air
        self.sediment = sediment
        self.threshold = threshold
        self.surfaceTracers = surfaceTracers
        self.Model = Model

    def _init_model(self):

        ErosionThreshold._init_model(self)
        SedimentationThreshold._init_model(self)

    def solve(self, dt):

        ErosionThreshold.solve(self, dt)
        SedimentationThreshold.solve(self, dt)



from __future__ import print_function,  absolute_import
import underworld as uw
import underworld.function as fn
import numpy as np
from mpi4py import MPI
from .LecodeIsostasy import LecodeIsostasy
from .scaling import nonDimensionalize as nd
from .scaling import UnitRegistry as u
from ._utils import Balanced_InflowOutflow
from ._utils import MovingWall
from ._boundary_conditions import BoundaryConditions

comm = MPI.COMM_WORLD
rank = comm.Get_rank()


class VelocityBCs(BoundaryConditions):
    """ Class to define the mechanical boundary conditions """

    def __init__(self, Model, left=None, right=None, top=None, bottom=None,
                 front=None, back=None, nodeSets=tuple(),
                 order_wall_conditions=None):
        """ Defines mechanical boundary conditions

        The type of conditions is determined through the units used do define
        the parameters:
            * Units of velocity ([length] / [time]) represent a kinematic
            condition (Dirichlet)
            * Units of stress / pressure ([force] / [area]) are set as
            stress condition (Neumann).

        parameters
        ----------

        Model: (UWGeodynamics.Model)
            An UWGeodynamics Model (See UWGeodynamics.Model)
        left:(tuple) with length 2 in 2D, and length 3 in 3D.
            Define mechanical conditions on the left side of the Model.
            Conditions are defined for each Model direction (x, y, [z])
        right:(tuple) with length 2 in 2D, and length 3 in 3D.
            Define mechanical conditions on the right side of the Model.
            Conditions are defined for each Model direction (x, y, [z])
        top:(tuple) with length 2 in 2D, and length 3 in 3D.
            Define mechanical conditions on the top side of the Model.
            Conditions are defined for each Model direction (x, y, [z])
        bottom:(tuple) with length 2 in 2D, and length 3 in 3D.
            Define mechanical conditions on the bottom side of the Model.
            Conditions are defined for each Model direction (x, y, [z])
        indexSets: [(condition, IndexSet)]
            List of node where to apply predefined velocities.

        Only valid for 3D Models:

        front:(tuple) with length 2 in 2D, and length 3 in 3D.
            Define mechanical conditions on the front side of the Model.
            Conditions are defined for each Model direction (x, y, [z])
        back:(tuple) with length 2 in 2D, and length 3 in 3D.
            Define mechanical conditions on the front side of the Model.
            Conditions are defined for each Model direction (x, y, [z])


        examples:
        ---------

        The following example defines a (2x1) meter Underworld model with
        freeslip conditions on all the sides.

        >>> import UWGeodynamics as GEO
        >>> u = GEO.u
        >>> Model = GEO.Model(elementRes=(64, 64),
                              minCoord=(-1. * u.meter, -50. * u.centimeter),
                              maxCoord=(1. * u.meter, 50. * u.centimeter))
        >>> Model.set_velocityBCs(left=[0, None], right=[0,None], top=[None,0],
                                  bottom=[None, 0])

        """

        super(VelocityBCs, self).__init__(Model, left, right, top, bottom,
                                          front, back, nodeSets, 
                                          order_wall_conditions)
        # Link Moving Walls
        for arg in [self.left, self.right, self.top, self.bottom, self.front,
                    self.back]:
            if isinstance(arg, MovingWall):
                arg.Model = self.Model

    def __getitem__(self, name):
        return self.__dict__[name]

    def _apply_conditions_nodes(self, condition, nodes):
        """ Apply condition to a set of nodes

        Parameters:
        -----------
            condition:
                velocity condition
            nodes:
                set of nodes

        """

        # Special case (Bottom LecodeIsostasy)
        if isinstance(condition, LecodeIsostasy):

            # Apply support condition
            self.Model._isostasy = self.bottom
            self.Model._isostasy.mesh = self.Model.mesh
            self.Model._isostasy.swarm = self.Model.swarm
            self.Model._isostasy._mesh_advector = self.Model._advector
            self.Model._isostasy.velocityField = self.Model.velocityField
            self.Model._isostasy.boundariesField = self.Model.boundariesField
            self.Model._isostasy.materialIndexField = self.Model.materialField
            self.Model._isostasy._densityFn = self.Model._densityFn
            vertical_walls_conditions = {
                "left": self.left,
                "right": self.right,
                "front": self.front,
                "back": self.back
            }
            self.Model._isostasy.vertical_walls_conditions = (
                vertical_walls_conditions)
            self._dirichlet_indices[-1] += self.Model.bottom_wall
            return

        if isinstance(condition, MovingWall):
            condition.wall = nodes
            set_, axis = condition.get_wall_indices()
            func = condition.velocityFn

            # Intersect of wall nodes and current local domain
            intersect = np.intersect1d(self.Model.mesh.data_nodegId, set_.data)
            ISet = uw.mesh.FeMesh_IndexSet(
                self.Model.mesh, topologicalIndex=0,
                size=self.Model.mesh.nodesGlobal,
                fromObject=intersect)

            for dim in range(self.Model.mesh.dim):
                if ISet.data.size > 0:
                    if (dim == axis):
                        self.Model.velocityField.data[ISet.data, dim] = (
                            func.evaluate(ISet)[:, 0])
                        self.Model.boundariesField.data[ISet.data, dim] = (
                            func.evaluate(ISet)[:, 0])
                        self._dirichlet_indices[dim] += ISet
                    else:
                        self.Model.velocityField.data[ISet.data, dim] = 0.
                        self.Model.boundariesField.data[ISet.data, dim] = 0.
                        self._dirichlet_indices[dim] += ISet

            return

        # Expect a list or tuple of dimension mesh.dim.
        # Check that the domain actually contains some boundary nodes
        # (nodes is not None)
        if isinstance(condition, (list, tuple)) and nodes.data.size > 0:
            for dim in range(self.Model.mesh.dim):

                if isinstance(condition[dim], fn.Function):
                    func = condition[dim]
                    self.Model.velocityField.data[nodes.data, dim] = (
                        func.evaluate(
                            self.Model.mesh.data[nodes.data])[:, dim])
                    self.Model.boundariesField.data[nodes.data, dim] = (
                        func.evaluate(
                            self.Model.mesh.data[nodes.data])[:, dim])
                    self._dirichlet_indices[dim] += nodes

                # User defined function
                if isinstance(condition[dim], (list, tuple)):
                    func = fn.branching.conditional(condition[dim])
                    self.Model.velocityField.data[nodes.data, dim] = (
                        func.evaluate(
                            self.Model.mesh.data[nodes.data])[:, 0])
                    self.Model.boundariesField.data[nodes.data, dim] = (
                        func.evaluate(
                            self.Model.mesh.data[nodes.data])[:, 0])
                    self._dirichlet_indices[dim] += nodes

                # Scalar condition
                if isinstance(condition[dim], (u.Quantity, float, int)):
                    self.Model.velocityField.data[nodes.data, dim] = (
                        nd(condition[dim]))
                    self.Model.boundariesField.data[nodes.data, dim] = (
                        nd(condition[dim]))
                    self._dirichlet_indices[dim] += nodes

                # Inflow Outflow
                if isinstance(condition[dim], Balanced_InflowOutflow):
                    obj = condition[dim]
                    obj.ynodes = self.Model.mesh.data[nodes.data, 1]
                    obj._get_side_flow()
                    self.Model.velocityField.data[nodes.data, dim] = (
                        obj._get_side_flow())
                    self.Model.boundariesField.data[nodes.data, dim] = (
                        obj._get_side_flow())
                    self._dirichlet_indices[dim] += nodes

                if isinstance(condition[dim], LecodeIsostasy):
                    # Apply support condition
                    if self.Model.mesh.dim - 1 != dim:
                        raise ValueError("""Can not apply LecodeIsostasy on that
                                         dimension""")

                    self.Model._isostasy = condition[dim]
                    self.Model._isostasy.mesh = self.Model.mesh
                    self.Model._isostasy.swarm = self.Model.swarm
                    self.Model._isostasy._mesh_advector = self.Model._advector
                    self.Model._isostasy.velocityField = self.Model.velocityField
                    self.Model._isostasy.boundariesField = self.Model.boundariesField
                    self.Model._isostasy.materialIndexField = self.Model.materialField
                    self.Model._isostasy._densityFn = self.Model._densityFn
                    vertical_walls_conditions = {
                        "left": self.left,
                        "right": self.right,
                        "front": self.front,
                        "back": self.back
                    }
                    self.Model._isostasy.vertical_walls_conditions = (
                        vertical_walls_conditions)
                    self._dirichlet_indices[dim] += nodes

        return

    def get_conditions(self):
        """ Get the mechanical boundary conditions

        Returns
        -------

        List of conditions as:
            [<underworld.conditions._conditions.DirichletCondition,
             <underworld.conditions._conditions.NeumannCondition]
        or
            [<underworld.conditions._conditions.DirichletCondition]

        """

        Model = self.Model

        # Reinitialise neumnann and dirichlet condition
        self._dirichlet_indices = []

        for _ in range(Model.mesh.dim):
            self._dirichlet_indices.append(Model.mesh.specialSets["Empty"])

        for set_ in self.order_wall_conditions:
            (condition, nodes) = self._wall_indexSets[set_]
            self._apply_conditions_nodes(condition, nodes)

        if self.nodeSets:
            for (condition, nodes) in self.nodeSets:
                self._apply_conditions_nodes(condition, nodes)

        dirichlet_conditions = None

        dirichlet_conditions = uw.conditions.DirichletCondition(
            variable=Model.velocityField,
            indexSetsPerDof=self._dirichlet_indices)

        return dirichlet_conditions


class StressBCs(BoundaryConditions):
    """ Class to define the stress boundary conditions """

    def __init__(self, Model, left=None, right=None, top=None, bottom=None,
                 front=None, back=None, nodeSets=tuple(),
                 order_wall_conditions=None):
        """ Defines stress boundary conditions

        parameters
        ----------

        Model: (UWGeodynamics.Model)
            An UWGeodynamics Model (See UWGeodynamics.Model)
        left:(tuple) with length 2 in 2D, and length 3 in 3D.
            Define conditions on the left side of the Model.
            Conditions are defined for each Model direction (x, y, [z])
        right:(tuple) with length 2 in 2D, and length 3 in 3D.
            Define conditions on the right side of the Model.
            Conditions are defined for each Model direction (x, y, [z])
        top:(tuple) with length 2 in 2D, and length 3 in 3D.
            Define conditions on the top side of the Model.
            Conditions are defined for each Model direction (x, y, [z])
        bottom:(tuple) with length 2 in 2D, and length 3 in 3D.
            Define conditions on the bottom side of the Model.
            Conditions are defined for each Model direction (x, y, [z])
        indexSets: [(condition, IndexSet)]
            List of node where to apply predefined velocities.

        Only valid for 3D Models:

        front:(tuple) with length 2 in 2D, and length 3 in 3D.
            Define mechanical conditions on the front side of the Model.
            Conditions are defined for each Model direction (x, y, [z])
        back:(tuple) with length 2 in 2D, and length 3 in 3D.
            Define mechanical conditions on the front side of the Model.
            Conditions are defined for each Model direction (x, y, [z])


        examples:
        ---------

        """
        super(StressBCs, self).__init__(Model, left, right, top, bottom,
                         front, back, nodeSets, order_wall_conditions)

    def __getitem__(self, name):
        return self.__dict__[name]

    def _apply_conditions_nodes(self, condition, nodes):
        """ Apply condition to a set of nodes

        Parameters:
        -----------
            condition:
                velocity condition
            nodes:
                set of nodes

        """

        if not nodes:
            return

        # Expect a list or tuple of dimension mesh.dim.
        # Check that the domain actually contains some boundary nodes
        # (nodes is not None)
        if isinstance(condition, (list, tuple)) and nodes.data.size > 0:
            for dim in range(self.Model.mesh.dim):

                # Scalar condition
                if isinstance(condition[dim], (u.Quantity, float, int)):
                    self.Model.tractionField.data[nodes.data, dim] = (
                        nd(condition[dim]))
                    self._neumann_indices[dim] += nodes

        return

    def get_conditions(self):
        """ Get the mechanical boundary conditions

        Returns
        -------

        List of conditions as:
            [<underworld.conditions._conditions.DirichletCondition,
             <underworld.conditions._conditions.NeumannCondition]
        or
            [<underworld.conditions._conditions.DirichletCondition]

        """

        Model = self.Model

        # Reinitialise neumnann condition
        self._neumann_indices = []

        for _ in range(Model.mesh.dim):
            self._neumann_indices.append(Model.mesh.specialSets["Empty"])

        for set_ in self.order_wall_conditions:
            (condition, nodes) = self._wall_indexSets[set_]
            self._apply_conditions_nodes(condition, nodes)

        if self.nodeSets:
            for (condition, nodes) in self.nodeSets:
                self._apply_conditions_nodes(condition, nodes)

        self.neumann_conditions = None
        _neumann_indices = []

        # Remove empty Sets
        for val in self._neumann_indices:
            if val.data.size > 0:
                _neumann_indices.append(val)
            else:
                _neumann_indices.append(None)
        self._neumann_indices = tuple(_neumann_indices)

        # Now we only create a Neumann condition if we have a stress condition
        # somewhere, on any of the procs.
        local_procs_has_neumann = np.zeros((uw.nProcs()))
        global_procs_has_neumann = np.zeros((uw.nProcs()))

        if self._neumann_indices != tuple([None for val in range(Model.mesh.dim)]):
            local_procs_has_neumann[uw.rank()] = 1

        comm.Allreduce(local_procs_has_neumann, global_procs_has_neumann)
        comm.Barrier()

        if any(global_procs_has_neumann):
            self.neumann_conditions = uw.conditions.NeumannCondition(
                fn_flux=Model.tractionField,
                variable=Model.velocityField,
                indexSetsPerDof=self._neumann_indices)

        return self.neumann_conditions

